# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tests of things related to the GraphBuilder class."""

import copy
import io
import logging
import textwrap
import unittest
from typing import Any

import lsst.pipe.base.automatic_connection_constants as acc
import lsst.utils.tests
from lsst.daf.butler import DataCoordinate, DatasetRef, DatasetType, DimensionUniverse, StorageClassFactory
from lsst.daf.butler.registry import MissingDatasetTypeError
from lsst.pipe.base.pipeline_graph import (
    ConnectionTypeConsistencyError,
    DuplicateOutputError,
    Edge,
    EdgesChangedError,
    IncompatibleDatasetTypeError,
    NodeKey,
    NodeType,
    PipelineGraph,
    PipelineGraphError,
    TaskImportMode,
    UnresolvedGraphError,
    visualization,
)
from lsst.pipe.base.tests.mocks import (
    DynamicConnectionConfig,
    DynamicTestPipelineTask,
    DynamicTestPipelineTaskConfig,
    get_mock_name,
)

_LOG = logging.getLogger(__name__)


class MockRegistry:
    """A test-utility stand-in for lsst.daf.butler.Registry that just knows
    how to get dataset types.
    """

    def __init__(self, dimensions: DimensionUniverse, dataset_types: dict[str, DatasetType]) -> None:
        self.dimensions = dimensions
        self._dataset_types = dataset_types

    def getDatasetType(self, name: str) -> DatasetType:
        try:
            return self._dataset_types[name]
        except KeyError:
            raise MissingDatasetTypeError(name) from None


class PipelineGraphTestCase(unittest.TestCase):
    """Tests for the `PipelineGraph` class.

    Tests for `PipelineGraph.resolve` are mostly in
    `PipelineGraphResolveTestCase` later in this file.
    """

    def setUp(self) -> None:
        # Simple test pipeline has two tasks, 'a' and 'b', with dataset types
        # 'input', 'intermediate', and 'output'.  There are no dimensions on
        # any of those.  We add tasks in reverse order to better test sorting.
        # There is one labeled task subset, 'only_b', with just 'b' in it.
        # We copy the configs so the originals (the instance attributes) can
        # be modified and reused after the ones passed in to the graph are
        # frozen.
        self.description = "A pipeline for PipelineGraph unit tests."
        self.graph = PipelineGraph()
        self.graph.description = self.description
        self.b_config = DynamicTestPipelineTaskConfig()
        self.b_config.init_inputs["in_schema"] = DynamicConnectionConfig(dataset_type_name="schema")
        self.b_config.inputs["input1"] = DynamicConnectionConfig(dataset_type_name="intermediate_1")
        self.b_config.outputs["output1"] = DynamicConnectionConfig(dataset_type_name="output_1")
        self.graph.add_task("b", DynamicTestPipelineTask, copy.deepcopy(self.b_config))
        self.a_config = DynamicTestPipelineTaskConfig()
        self.a_config.init_outputs["out_schema"] = DynamicConnectionConfig(dataset_type_name="schema")
        self.a_config.inputs["input1"] = DynamicConnectionConfig(dataset_type_name="input_1")
        self.a_config.outputs["output1"] = DynamicConnectionConfig(dataset_type_name="intermediate_1")
        self.graph.add_task("a", DynamicTestPipelineTask, copy.deepcopy(self.a_config))
        self.graph.add_task_subset("only_b", ["b"])
        self.subset_description = "A subset with only task B in it."
        self.graph.task_subsets["only_b"].description = self.subset_description
        self.dimensions = DimensionUniverse()
        self.maxDiff = None

    def test_unresolved_accessors(self) -> None:
        """Test attribute accessors, iteration, and simple methods on a graph
        that has not had `PipelineGraph.resolve` called on it.
        """
        self.check_base_accessors(self.graph)
        self.assertEqual(
            repr(self.graph.tasks["a"]), "a (lsst.pipe.base.tests.mocks.DynamicTestPipelineTask)"
        )

    def test_sorting(self) -> None:
        """Test sort methods on PipelineGraph."""
        self.assertFalse(self.graph.has_been_sorted)
        self.assertFalse(self.graph.is_sorted)
        self.graph.sort()
        self.check_sorted(self.graph)

    def test_unresolved_xgraph_export(self) -> None:
        """Test exporting an unresolved PipelineGraph to networkx in various
        ways.
        """
        self.check_make_xgraph(self.graph, resolved=False)
        self.check_make_bipartite_xgraph(self.graph, resolved=False)
        self.check_make_task_xgraph(self.graph, resolved=False)
        self.check_make_dataset_type_xgraph(self.graph, resolved=False)

    def test_unresolved_stream_io(self) -> None:
        """Test round-tripping an unresolved PipelineGraph through in-memory
        serialization.
        """
        stream = io.BytesIO()
        self.graph._write_stream(stream)
        stream.seek(0)
        roundtripped = PipelineGraph._read_stream(stream)
        self.check_make_xgraph(roundtripped, resolved=False)

    def test_unresolved_file_io(self) -> None:
        """Test round-tripping an unresolved PipelineGraph through file
        serialization.
        """
        with lsst.utils.tests.getTempFilePath(".json.gz") as filename:
            self.graph._write_uri(filename)
            roundtripped = PipelineGraph._read_uri(filename)
        self.check_make_xgraph(roundtripped, resolved=False)

    def test_unresolved_deferred_import_io(self) -> None:
        """Test round-tripping an unresolved PipelineGraph through
        serialization, without immediately importing tasks on read.
        """
        stream = io.BytesIO()
        self.graph._write_stream(stream)
        stream.seek(0)
        roundtripped = PipelineGraph._read_stream(stream, import_mode=TaskImportMode.DO_NOT_IMPORT)
        self.check_make_xgraph(roundtripped, resolved=False, imported_and_configured=False)
        # Check that we can still resolve the graph without importing tasks.
        roundtripped.resolve(MockRegistry(self.dimensions, {}))
        self.check_make_xgraph(roundtripped, resolved=True, imported_and_configured=False)
        roundtripped._import_and_configure(TaskImportMode.ASSUME_CONSISTENT_EDGES)
        self.check_make_xgraph(roundtripped, resolved=True, imported_and_configured=True)

    def test_resolved_accessors(self) -> None:
        """Test attribute accessors, iteration, and simple methods on a graph
        that has had `PipelineGraph.resolve` called on it.

        This includes the accessors available on unresolved graphs as well as
        new ones, and we expect the resolved graph to be sorted as well.
        """
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        self.check_base_accessors(self.graph)
        self.check_sorted(self.graph)
        self.assertEqual(
            repr(self.graph.tasks["a"]), "a (lsst.pipe.base.tests.mocks.DynamicTestPipelineTask, {})"
        )
        self.assertEqual(self.graph.tasks["a"].dimensions, self.dimensions.empty)
        self.assertEqual(repr(self.graph.dataset_types["input_1"]), "input_1 (_mock_StructuredDataDict, {})")
        self.assertEqual(self.graph.dataset_types["input_1"].key, NodeKey(NodeType.DATASET_TYPE, "input_1"))
        self.assertEqual(self.graph.dataset_types["input_1"].dimensions, self.dimensions.empty)
        self.assertEqual(self.graph.dataset_types["input_1"].storage_class_name, "_mock_StructuredDataDict")
        self.assertEqual(self.graph.dataset_types["input_1"].storage_class.name, "_mock_StructuredDataDict")

    def test_resolved_xgraph_export(self) -> None:
        """Test exporting a resolved PipelineGraph to networkx in various
        ways.
        """
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        self.check_make_xgraph(self.graph, resolved=True)
        self.check_make_bipartite_xgraph(self.graph, resolved=True)
        self.check_make_task_xgraph(self.graph, resolved=True)
        self.check_make_dataset_type_xgraph(self.graph, resolved=True)

    def test_resolved_stream_io(self) -> None:
        """Test round-tripping a resolved PipelineGraph through in-memory
        serialization.
        """
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        stream = io.BytesIO()
        self.graph._write_stream(stream)
        stream.seek(0)
        roundtripped = PipelineGraph._read_stream(stream)
        self.check_make_xgraph(roundtripped, resolved=True)

    def test_resolved_file_io(self) -> None:
        """Test round-tripping a resolved PipelineGraph through file
        serialization.
        """
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        with lsst.utils.tests.getTempFilePath(".json.gz") as filename:
            self.graph._write_uri(filename)
            roundtripped = PipelineGraph._read_uri(filename)
        self.check_make_xgraph(roundtripped, resolved=True)

    def test_resolved_deferred_import_io(self) -> None:
        """Test round-tripping a resolved PipelineGraph through serialization,
        without immediately importing tasks on read.
        """
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        stream = io.BytesIO()
        self.graph._write_stream(stream)
        stream.seek(0)
        roundtripped = PipelineGraph._read_stream(stream, import_mode=TaskImportMode.DO_NOT_IMPORT)
        self.check_make_xgraph(roundtripped, resolved=True, imported_and_configured=False)
        roundtripped._import_and_configure(TaskImportMode.REQUIRE_CONSISTENT_EDGES)
        self.check_make_xgraph(roundtripped, resolved=True, imported_and_configured=True)

    def test_unresolved_copies(self) -> None:
        """Test making copies of an unresolved PipelineGraph."""
        copy1 = self.graph.copy()
        self.assertIsNot(copy1, self.graph)
        self.check_make_xgraph(copy1, resolved=False)
        copy2 = copy.copy(self.graph)
        self.assertIsNot(copy2, self.graph)
        self.check_make_xgraph(copy2, resolved=False)
        copy3 = copy.deepcopy(self.graph)
        self.assertIsNot(copy3, self.graph)
        self.check_make_xgraph(copy3, resolved=False)

    def test_resolved_copies(self) -> None:
        """Test making copies of a resolved PipelineGraph."""
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        copy1 = self.graph.copy()
        self.assertIsNot(copy1, self.graph)
        self.check_make_xgraph(copy1, resolved=True)
        copy2 = copy.copy(self.graph)
        self.assertIsNot(copy2, self.graph)
        self.check_make_xgraph(copy2, resolved=True)
        copy3 = copy.deepcopy(self.graph)
        self.assertIsNot(copy3, self.graph)
        self.check_make_xgraph(copy3, resolved=True)

    def check_base_accessors(self, graph: PipelineGraph) -> None:
        """Run parameterized tests that check attribute access, iteration, and
        simple methods.

        The given graph must be unchanged from the one defined in `setUp`,
        other than sorting.
        """
        self.assertEqual(graph.description, self.description)
        self.assertEqual(graph.tasks.keys(), {"a", "b"})
        self.assertEqual(
            graph.dataset_types.keys(),
            {
                "schema",
                "input_1",
                "intermediate_1",
                "output_1",
                "a_config",
                "a_log",
                "a_metadata",
                "b_config",
                "b_log",
                "b_metadata",
            },
        )
        self.assertEqual(graph.task_subsets.keys(), {"only_b"})
        self.assertEqual(
            {edge.nodes + (repr(edge),) for edge in graph.iter_edges(init=False)},
            {
                (
                    NodeKey(NodeType.DATASET_TYPE, "input_1"),
                    NodeKey(NodeType.TASK, "a"),
                    "input_1 -> a (input1)",
                ),
                (
                    NodeKey(NodeType.TASK, "a"),
                    NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                    "a -> intermediate_1 (output1)",
                ),
                (
                    NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                    NodeKey(NodeType.TASK, "b"),
                    "intermediate_1 -> b (input1)",
                ),
                (
                    NodeKey(NodeType.TASK, "b"),
                    NodeKey(NodeType.DATASET_TYPE, "output_1"),
                    "b -> output_1 (output1)",
                ),
                (NodeKey(NodeType.TASK, "a"), NodeKey(NodeType.DATASET_TYPE, "a_log"), "a -> a_log (_log)"),
                (
                    NodeKey(NodeType.TASK, "a"),
                    NodeKey(NodeType.DATASET_TYPE, "a_metadata"),
                    "a -> a_metadata (_metadata)",
                ),
                (NodeKey(NodeType.TASK, "b"), NodeKey(NodeType.DATASET_TYPE, "b_log"), "b -> b_log (_log)"),
                (
                    NodeKey(NodeType.TASK, "b"),
                    NodeKey(NodeType.DATASET_TYPE, "b_metadata"),
                    "b -> b_metadata (_metadata)",
                ),
            },
        )
        self.assertEqual(
            {edge.nodes + (repr(edge),) for edge in graph.iter_edges(init=True)},
            {
                (
                    NodeKey(NodeType.TASK_INIT, "a"),
                    NodeKey(NodeType.DATASET_TYPE, "schema"),
                    "a -> schema (out_schema)",
                ),
                (
                    NodeKey(NodeType.DATASET_TYPE, "schema"),
                    NodeKey(NodeType.TASK_INIT, "b"),
                    "schema -> b (in_schema)",
                ),
                (
                    NodeKey(NodeType.TASK_INIT, "a"),
                    NodeKey(NodeType.DATASET_TYPE, "a_config"),
                    "a -> a_config (_config)",
                ),
                (
                    NodeKey(NodeType.TASK_INIT, "b"),
                    NodeKey(NodeType.DATASET_TYPE, "b_config"),
                    "b -> b_config (_config)",
                ),
            },
        )
        self.assertEqual(
            {(node_type, name) for node_type, name, _ in graph.iter_nodes()},
            {
                NodeKey(NodeType.TASK, "a"),
                NodeKey(NodeType.TASK, "b"),
                NodeKey(NodeType.TASK_INIT, "a"),
                NodeKey(NodeType.TASK_INIT, "b"),
                NodeKey(NodeType.DATASET_TYPE, "schema"),
                NodeKey(NodeType.DATASET_TYPE, "input_1"),
                NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                NodeKey(NodeType.DATASET_TYPE, "output_1"),
                NodeKey(NodeType.DATASET_TYPE, "a_config"),
                NodeKey(NodeType.DATASET_TYPE, "a_log"),
                NodeKey(NodeType.DATASET_TYPE, "a_metadata"),
                NodeKey(NodeType.DATASET_TYPE, "b_config"),
                NodeKey(NodeType.DATASET_TYPE, "b_log"),
                NodeKey(NodeType.DATASET_TYPE, "b_metadata"),
            },
        )
        self.assertEqual({name for name, _ in graph.iter_overall_inputs()}, {"input_1"})
        self.assertEqual({edge.task_label for edge in graph.consuming_edges_of("input_1")}, {"a"})
        self.assertEqual({edge.task_label for edge in graph.consuming_edges_of("intermediate_1")}, {"b"})
        self.assertEqual({edge.task_label for edge in graph.consuming_edges_of("output_1")}, set())
        self.assertEqual({node.label for node in graph.consumers_of("input_1")}, {"a"})
        self.assertEqual({node.label for node in graph.consumers_of("intermediate_1")}, {"b"})
        self.assertEqual({node.label for node in graph.consumers_of("output_1")}, set())

        self.assertIsNone(graph.producing_edge_of("input_1"))
        self.assertEqual(graph.producing_edge_of("intermediate_1").task_label, "a")
        self.assertEqual(graph.producing_edge_of("output_1").task_label, "b")
        self.assertIsNone(graph.producer_of("input_1"))
        self.assertEqual(graph.producer_of("intermediate_1").label, "a")
        self.assertEqual(graph.producer_of("output_1").label, "b")

        self.assertEqual(graph.inputs_of("a").keys(), {"input_1"})
        self.assertEqual(graph.inputs_of("b").keys(), {"intermediate_1"})
        self.assertEqual(graph.inputs_of("a", init=True).keys(), set())
        self.assertEqual(graph.inputs_of("b", init=True).keys(), {"schema"})
        self.assertEqual(graph.outputs_of("a").keys(), {"intermediate_1", "a_log", "a_metadata"})
        self.assertEqual(graph.outputs_of("b").keys(), {"output_1", "b_log", "b_metadata"})
        self.assertEqual(
            graph.outputs_of("a", include_automatic_connections=False).keys(), {"intermediate_1"}
        )
        self.assertEqual(graph.outputs_of("b", include_automatic_connections=False).keys(), {"output_1"})
        self.assertEqual(graph.outputs_of("a", init=True).keys(), {"schema", "a_config"})
        self.assertEqual(
            graph.outputs_of("a", init=True, include_automatic_connections=False).keys(), {"schema"}
        )
        self.assertEqual(graph.outputs_of("b", init=True).keys(), {"b_config"})
        self.assertEqual(graph.outputs_of("b", init=True, include_automatic_connections=False).keys(), set())

        self.assertTrue(repr(self.graph).startswith(f"PipelineGraph({self.description!r}, tasks="))
        self.assertEqual(
            repr(graph.task_subsets["only_b"]), f"only_b: {self.subset_description!r}, tasks={{b}}"
        )

    def check_sorted(self, graph: PipelineGraph) -> None:
        """Run a battery of tests on a PipelineGraph that must be
        deterministically sorted.

        The given graph must be unchanged from the one defined in `setUp`,
        other than sorting.
        """
        self.assertTrue(graph.has_been_sorted)
        self.assertTrue(graph.is_sorted)
        self.assertEqual(
            [(node_type, name) for node_type, name, _ in graph.iter_nodes()],
            [
                # We only advertise that the order is topological and
                # deterministic, so this test is slightly over-specified; there
                # are other orders that are consistent with our guarantees.
                NodeKey(NodeType.DATASET_TYPE, "input_1"),
                NodeKey(NodeType.TASK_INIT, "a"),
                NodeKey(NodeType.DATASET_TYPE, "a_config"),
                NodeKey(NodeType.DATASET_TYPE, "schema"),
                NodeKey(NodeType.TASK_INIT, "b"),
                NodeKey(NodeType.DATASET_TYPE, "b_config"),
                NodeKey(NodeType.TASK, "a"),
                NodeKey(NodeType.DATASET_TYPE, "a_log"),
                NodeKey(NodeType.DATASET_TYPE, "a_metadata"),
                NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                NodeKey(NodeType.TASK, "b"),
                NodeKey(NodeType.DATASET_TYPE, "b_log"),
                NodeKey(NodeType.DATASET_TYPE, "b_metadata"),
                NodeKey(NodeType.DATASET_TYPE, "output_1"),
            ],
        )
        # Most users should only care that the tasks and dataset types are
        # topologically sorted.
        self.assertEqual(list(graph.tasks), ["a", "b"])
        self.assertEqual(
            list(graph.dataset_types),
            [
                "input_1",
                "a_config",
                "schema",
                "b_config",
                "a_log",
                "a_metadata",
                "intermediate_1",
                "b_log",
                "b_metadata",
                "output_1",
            ],
        )
        # __str__ and __repr__ of course work on unsorted mapping views, too,
        # but the order of elements is then nondeterministic and hard to test.
        self.assertEqual(repr(self.graph.tasks), "TaskMappingView({a, b})")
        self.assertEqual(
            repr(self.graph.dataset_types),
            (
                "DatasetTypeMappingView({input_1, a_config, schema, b_config, a_log, a_metadata, "
                "intermediate_1, b_log, b_metadata, output_1})"
            ),
        )

    def check_make_xgraph(
        self, graph: PipelineGraph, resolved: bool, imported_and_configured: bool = True
    ) -> None:
        """Check that the given graph exports as expected to networkx.

        The given graph must be unchanged from the one defined in `setUp`,
        other than being resolved (if ``resolved=True``) or round-tripped
        through serialization without tasks being imported (if
        ``imported_and_configured=False``).
        """
        xgraph = graph.make_xgraph()
        expected_edges = (
            {edge.key for edge in graph.iter_edges()}
            | {edge.key for edge in graph.iter_edges(init=True)}
            | {
                (NodeKey(NodeType.TASK_INIT, "a"), NodeKey(NodeType.TASK, "a"), Edge.INIT_TO_TASK_NAME),
                (NodeKey(NodeType.TASK_INIT, "b"), NodeKey(NodeType.TASK, "b"), Edge.INIT_TO_TASK_NAME),
            }
        )
        test_edges = set(xgraph.edges)
        self.assertEqual(test_edges, expected_edges)
        expected_nodes = {
            NodeKey(NodeType.TASK_INIT, "a"): self.get_expected_task_init_node(
                "a", resolved, imported_and_configured=imported_and_configured
            ),
            NodeKey(NodeType.TASK, "a"): self.get_expected_task_node(
                "a", resolved, imported_and_configured=imported_and_configured
            ),
            NodeKey(NodeType.TASK_INIT, "b"): self.get_expected_task_init_node(
                "b", resolved, imported_and_configured=imported_and_configured
            ),
            NodeKey(NodeType.TASK, "b"): self.get_expected_task_node(
                "b", resolved, imported_and_configured=imported_and_configured
            ),
            NodeKey(NodeType.DATASET_TYPE, "a_config"): self.get_expected_config_node("a", resolved),
            NodeKey(NodeType.DATASET_TYPE, "b_config"): self.get_expected_config_node("b", resolved),
            NodeKey(NodeType.DATASET_TYPE, "a_log"): self.get_expected_log_node("a", resolved),
            NodeKey(NodeType.DATASET_TYPE, "b_log"): self.get_expected_log_node("b", resolved),
            NodeKey(NodeType.DATASET_TYPE, "a_metadata"): self.get_expected_metadata_node("a", resolved),
            NodeKey(NodeType.DATASET_TYPE, "b_metadata"): self.get_expected_metadata_node("b", resolved),
            NodeKey(NodeType.DATASET_TYPE, "schema"): self.get_expected_connection_node(
                "schema", resolved, is_initial_query_constraint=False
            ),
            NodeKey(NodeType.DATASET_TYPE, "input_1"): self.get_expected_connection_node(
                "input_1", resolved, is_initial_query_constraint=True
            ),
            NodeKey(NodeType.DATASET_TYPE, "intermediate_1"): self.get_expected_connection_node(
                "intermediate_1", resolved, is_initial_query_constraint=False
            ),
            NodeKey(NodeType.DATASET_TYPE, "output_1"): self.get_expected_connection_node(
                "output_1", resolved, is_initial_query_constraint=False
            ),
        }
        test_nodes = dict(xgraph.nodes.items())
        self.assertEqual(set(test_nodes.keys()), set(expected_nodes.keys()))
        for key, expected_node in expected_nodes.items():
            test_node = test_nodes[key]
            self.assertEqual(expected_node, test_node, key)

    def check_make_bipartite_xgraph(self, graph: PipelineGraph, resolved: bool) -> None:
        """Check that the given graph's init-only or runtime subset exports as
        expected to networkx.

        The given graph must be unchanged from the one defined in `setUp`,
        other than being resolved (if ``resolved=True``).
        """
        run_xgraph = graph.make_bipartite_xgraph()
        self.assertEqual(set(run_xgraph.edges), {edge.key for edge in graph.iter_edges()})
        self.assertEqual(
            dict(run_xgraph.nodes.items()),
            {
                NodeKey(NodeType.TASK, "a"): self.get_expected_task_node("a", resolved),
                NodeKey(NodeType.TASK, "b"): self.get_expected_task_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "a_log"): self.get_expected_log_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_log"): self.get_expected_log_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "a_metadata"): self.get_expected_metadata_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_metadata"): self.get_expected_metadata_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "input_1"): self.get_expected_connection_node(
                    "input_1", resolved, is_initial_query_constraint=True
                ),
                NodeKey(NodeType.DATASET_TYPE, "intermediate_1"): self.get_expected_connection_node(
                    "intermediate_1", resolved, is_initial_query_constraint=False
                ),
                NodeKey(NodeType.DATASET_TYPE, "output_1"): self.get_expected_connection_node(
                    "output_1", resolved, is_initial_query_constraint=False
                ),
            },
        )
        init_xgraph = graph.make_bipartite_xgraph(
            init=True,
        )
        self.assertEqual(set(init_xgraph.edges), {edge.key for edge in graph.iter_edges(init=True)})
        self.assertEqual(
            dict(init_xgraph.nodes.items()),
            {
                NodeKey(NodeType.TASK_INIT, "a"): self.get_expected_task_init_node("a", resolved),
                NodeKey(NodeType.TASK_INIT, "b"): self.get_expected_task_init_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "schema"): self.get_expected_connection_node(
                    "schema", resolved, is_initial_query_constraint=False
                ),
                NodeKey(NodeType.DATASET_TYPE, "a_config"): self.get_expected_config_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_config"): self.get_expected_config_node("b", resolved),
            },
        )

    def check_make_task_xgraph(self, graph: PipelineGraph, resolved: bool) -> None:
        """Check that the given graph's task-only projection exports as
        expected to networkx.

        The given graph must be unchanged from the one defined in `setUp`,
        other than being resolved (if ``resolved=True``).
        """
        run_xgraph = graph.make_task_xgraph()
        self.assertEqual(set(run_xgraph.edges), {(NodeKey(NodeType.TASK, "a"), NodeKey(NodeType.TASK, "b"))})
        self.assertEqual(
            dict(run_xgraph.nodes.items()),
            {
                NodeKey(NodeType.TASK, "a"): self.get_expected_task_node("a", resolved),
                NodeKey(NodeType.TASK, "b"): self.get_expected_task_node("b", resolved),
            },
        )
        init_xgraph = graph.make_task_xgraph(
            init=True,
        )
        self.assertEqual(
            set(init_xgraph.edges),
            {(NodeKey(NodeType.TASK_INIT, "a"), NodeKey(NodeType.TASK_INIT, "b"))},
        )
        self.assertEqual(
            dict(init_xgraph.nodes.items()),
            {
                NodeKey(NodeType.TASK_INIT, "a"): self.get_expected_task_init_node("a", resolved),
                NodeKey(NodeType.TASK_INIT, "b"): self.get_expected_task_init_node("b", resolved),
            },
        )

    def check_make_dataset_type_xgraph(self, graph: PipelineGraph, resolved: bool) -> None:
        """Check that the given graph's dataset-type-only projection exports as
        expected to networkx.

        The given graph must be unchanged from the one defined in `setUp`,
        other than being resolved (if ``resolved=True``).
        """
        run_xgraph = graph.make_dataset_type_xgraph()
        self.assertEqual(
            set(run_xgraph.edges),
            {
                (NodeKey(NodeType.DATASET_TYPE, "input_1"), NodeKey(NodeType.DATASET_TYPE, "intermediate_1")),
                (NodeKey(NodeType.DATASET_TYPE, "input_1"), NodeKey(NodeType.DATASET_TYPE, "a_log")),
                (NodeKey(NodeType.DATASET_TYPE, "input_1"), NodeKey(NodeType.DATASET_TYPE, "a_metadata")),
                (
                    NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                    NodeKey(NodeType.DATASET_TYPE, "output_1"),
                ),
                (NodeKey(NodeType.DATASET_TYPE, "intermediate_1"), NodeKey(NodeType.DATASET_TYPE, "b_log")),
                (
                    NodeKey(NodeType.DATASET_TYPE, "intermediate_1"),
                    NodeKey(NodeType.DATASET_TYPE, "b_metadata"),
                ),
            },
        )
        self.assertEqual(
            dict(run_xgraph.nodes.items()),
            {
                NodeKey(NodeType.DATASET_TYPE, "a_log"): self.get_expected_log_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_log"): self.get_expected_log_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "a_metadata"): self.get_expected_metadata_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_metadata"): self.get_expected_metadata_node("b", resolved),
                NodeKey(NodeType.DATASET_TYPE, "input_1"): self.get_expected_connection_node(
                    "input_1", resolved, is_initial_query_constraint=True
                ),
                NodeKey(NodeType.DATASET_TYPE, "intermediate_1"): self.get_expected_connection_node(
                    "intermediate_1", resolved, is_initial_query_constraint=False
                ),
                NodeKey(NodeType.DATASET_TYPE, "output_1"): self.get_expected_connection_node(
                    "output_1", resolved, is_initial_query_constraint=False
                ),
            },
        )
        init_xgraph = graph.make_dataset_type_xgraph(init=True)
        self.assertEqual(
            set(init_xgraph.edges),
            {(NodeKey(NodeType.DATASET_TYPE, "schema"), NodeKey(NodeType.DATASET_TYPE, "b_config"))},
        )
        self.assertEqual(
            dict(init_xgraph.nodes.items()),
            {
                NodeKey(NodeType.DATASET_TYPE, "schema"): self.get_expected_connection_node(
                    "schema", resolved, is_initial_query_constraint=False
                ),
                NodeKey(NodeType.DATASET_TYPE, "a_config"): self.get_expected_config_node("a", resolved),
                NodeKey(NodeType.DATASET_TYPE, "b_config"): self.get_expected_config_node("b", resolved),
            },
        )

    def get_expected_task_node(
        self, label: str, resolved: bool, imported_and_configured: bool = True
    ) -> dict[str, Any]:
        """Construct a networkx-export task node for comparison."""
        result = self.get_expected_task_init_node(
            label, resolved, imported_and_configured=imported_and_configured
        )
        if resolved:
            result["dimensions"] = self.dimensions.empty
        result["raw_dimensions"] = frozenset()
        return result

    def get_expected_task_init_node(
        self, label: str, resolved: bool, imported_and_configured: bool = True
    ) -> dict[str, Any]:
        """Construct a networkx-export task init for comparison."""
        result = {
            "task_class_name": "lsst.pipe.base.tests.mocks.DynamicTestPipelineTask",
            "bipartite": 1,
        }
        if imported_and_configured:
            result["task_class"] = DynamicTestPipelineTask
            result["config"] = getattr(self, f"{label}_config")
        return result

    def get_expected_config_node(self, label: str, resolved: bool) -> dict[str, Any]:
        """Construct a networkx-export init-output config dataset type node for
        comparison.
        """
        if not resolved:
            return {"bipartite": 0}
        else:
            return {
                "dataset_type": DatasetType(
                    acc.CONFIG_INIT_OUTPUT_TEMPLATE.format(label=label),
                    self.dimensions.empty,
                    acc.CONFIG_INIT_OUTPUT_STORAGE_CLASS,
                ),
                "is_initial_query_constraint": False,
                "is_prerequisite": False,
                "dimensions": self.dimensions.empty,
                "storage_class_name": acc.CONFIG_INIT_OUTPUT_STORAGE_CLASS,
                "bipartite": 0,
            }

    def get_expected_log_node(self, label: str, resolved: bool) -> dict[str, Any]:
        """Construct a networkx-export output log dataset type node for
        comparison.
        """
        if not resolved:
            return {"bipartite": 0}
        else:
            return {
                "dataset_type": DatasetType(
                    acc.LOG_OUTPUT_TEMPLATE.format(label=label),
                    self.dimensions.empty,
                    acc.LOG_OUTPUT_STORAGE_CLASS,
                ),
                "is_initial_query_constraint": False,
                "is_prerequisite": False,
                "dimensions": self.dimensions.empty,
                "storage_class_name": acc.LOG_OUTPUT_STORAGE_CLASS,
                "bipartite": 0,
            }

    def get_expected_metadata_node(self, label: str, resolved: bool) -> dict[str, Any]:
        """Construct a networkx-export output metadata dataset type node for
        comparison.
        """
        if not resolved:
            return {"bipartite": 0}
        else:
            return {
                "dataset_type": DatasetType(
                    acc.METADATA_OUTPUT_TEMPLATE.format(label=label),
                    self.dimensions.empty,
                    acc.METADATA_OUTPUT_STORAGE_CLASS,
                ),
                "is_initial_query_constraint": False,
                "is_prerequisite": False,
                "dimensions": self.dimensions.empty,
                "storage_class_name": acc.METADATA_OUTPUT_STORAGE_CLASS,
                "bipartite": 0,
            }

    def get_expected_connection_node(
        self, name: str, resolved: bool, *, is_initial_query_constraint: bool
    ) -> dict[str, Any]:
        """Construct a networkx-export dataset type node for comparison."""
        if not resolved:
            return {"bipartite": 0}
        else:
            return {
                "dataset_type": DatasetType(
                    name,
                    self.dimensions.empty,
                    get_mock_name("StructuredDataDict"),
                ),
                "is_initial_query_constraint": is_initial_query_constraint,
                "is_prerequisite": False,
                "dimensions": self.dimensions.empty,
                "storage_class_name": get_mock_name("StructuredDataDict"),
                "bipartite": 0,
            }

    def test_construct_with_data_coordinate(self) -> None:
        """Test constructing a graph with a DataCoordinate.

        Since this creates a graph with DimensionUniverse, all tasks added to
        it should have resolved dimensions, but not (yet) resolved dataset
        types.  We use that to test a few other operations in that state.
        """
        data_id = DataCoordinate.standardize(instrument="I", universe=self.dimensions)
        graph = PipelineGraph(data_id=data_id)
        self.assertEqual(graph.universe, self.dimensions)
        self.assertEqual(graph.data_id, data_id)
        graph.add_task("b1", DynamicTestPipelineTask, self.b_config)
        self.assertEqual(graph.tasks["b1"].dimensions, self.dimensions.empty)
        # Still can't group by dimensions, because the dataset types aren't
        # resolved.
        with self.assertRaises(UnresolvedGraphError):
            graph.group_by_dimensions()
        # Transferring a node from this graph to ``self.graph`` should
        # unresolve the dimensions.
        self.graph.add_task_nodes([graph.tasks["b1"]])
        self.assertIsNot(self.graph.tasks["b1"], graph.tasks["b1"])
        self.assertFalse(self.graph.tasks["b1"].has_resolved_dimensions)
        # Do the opposite transfer, which should resolve dimensions.
        graph.add_task_nodes([self.graph.tasks["a"]])
        self.assertIsNot(self.graph.tasks["a"], graph.tasks["a"])
        self.assertTrue(graph.tasks["a"].has_resolved_dimensions)

    def test_group_by_dimensions(self) -> None:
        """Test PipelineGraph.group_by_dimensions."""
        with self.assertRaises(UnresolvedGraphError):
            self.graph.group_by_dimensions()
        self.a_config.dimensions = ["visit"]
        self.a_config.outputs["output1"].dimensions = ["visit"]
        self.a_config.prerequisite_inputs["prereq1"] = DynamicConnectionConfig(
            dataset_type_name="prereq_1",
            multiple=True,
            dimensions=["htm7"],
            is_calibration=True,
        )
        self.b_config.dimensions = ["htm7"]
        self.b_config.inputs["input1"].dimensions = ["visit"]
        self.b_config.inputs["input1"].multiple = True
        self.b_config.outputs["output1"].dimensions = ["htm7"]
        self.graph.reconfigure_tasks(a=self.a_config, b=self.b_config)
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        visit_dims = self.dimensions.conform(["visit"])
        htm7_dims = self.dimensions.conform(["htm7"])
        expected = {
            self.dimensions.empty.as_group(): (
                {},
                {
                    "schema": self.graph.dataset_types["schema"],
                    "input_1": self.graph.dataset_types["input_1"],
                    "a_config": self.graph.dataset_types["a_config"],
                    "b_config": self.graph.dataset_types["b_config"],
                },
            ),
            visit_dims: (
                {"a": self.graph.tasks["a"]},
                {
                    "a_log": self.graph.dataset_types["a_log"],
                    "a_metadata": self.graph.dataset_types["a_metadata"],
                    "intermediate_1": self.graph.dataset_types["intermediate_1"],
                },
            ),
            htm7_dims: (
                {"b": self.graph.tasks["b"]},
                {
                    "b_log": self.graph.dataset_types["b_log"],
                    "b_metadata": self.graph.dataset_types["b_metadata"],
                    "output_1": self.graph.dataset_types["output_1"],
                },
            ),
        }
        self.assertEqual(self.graph.group_by_dimensions(), expected)
        expected[htm7_dims][1]["prereq_1"] = self.graph.dataset_types["prereq_1"]
        self.assertEqual(self.graph.group_by_dimensions(prerequisites=True), expected)

    def test_add_and_remove(self) -> None:
        """Tests for adding and removing tasks and task subsets from a
        PipelineGraph.
        """
        # Can't remove a task while it's still in a subset.
        with self.assertRaises(PipelineGraphError):
            self.graph.remove_tasks(["b"], drop_from_subsets=False)
        # ...unless you remove the subset.
        self.graph.remove_task_subset("only_b")
        self.assertFalse(self.graph.task_subsets)
        ((b, referencing_subsets),) = self.graph.remove_tasks(["b"], drop_from_subsets=False)
        self.assertFalse(referencing_subsets)
        self.assertEqual(self.graph.tasks.keys(), {"a"})
        # Add that task back in.
        self.graph.add_task_nodes([b])
        self.assertEqual(self.graph.tasks.keys(), {"a", "b"})
        # Add the subset back in.
        self.graph.add_task_subset("only_b", {"b"})
        self.assertEqual(self.graph.task_subsets.keys(), {"only_b"})
        # Resolve the graph's dataset types and task dimensions.
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        self.assertTrue(self.graph.dataset_types.is_resolved("input_1"))
        self.assertTrue(self.graph.dataset_types.is_resolved("output_1"))
        self.assertTrue(self.graph.dataset_types.is_resolved("schema"))
        self.assertTrue(self.graph.dataset_types.is_resolved("intermediate_1"))
        # Remove the task while removing it from the subset automatically. This
        # should also unresolve (only) the referenced dataset types and drop
        # any datasets no longer attached to any task.
        self.assertEqual(self.graph.tasks.keys(), {"a", "b"})
        ((b, referencing_subsets),) = self.graph.remove_tasks(["b"], drop_from_subsets=True)
        self.assertEqual(referencing_subsets, {"only_b"})
        self.assertEqual(self.graph.tasks.keys(), {"a"})
        self.assertTrue(self.graph.dataset_types.is_resolved("input_1"))
        self.assertNotIn("output1", self.graph.dataset_types)
        self.assertFalse(self.graph.dataset_types.is_resolved("schema"))
        self.assertFalse(self.graph.dataset_types.is_resolved("intermediate_1"))

    def test_reconfigure(self) -> None:
        """Tests for PipelineGraph.reconfigure."""
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        self.b_config.outputs["output1"].storage_class = "TaskMetadata"
        with self.assertRaises(ValueError):
            # Can't check and assume together.
            self.graph.reconfigure_tasks(
                b=self.b_config, assume_edges_unchanged=True, check_edges_unchanged=True
            )
        # Check that graph is unchanged after error.
        self.check_base_accessors(self.graph)
        with self.assertRaises(EdgesChangedError):
            self.graph.reconfigure_tasks(b=self.b_config, check_edges_unchanged=True)
        self.check_base_accessors(self.graph)
        # Make a change that does affect edges; this will unresolve most
        # dataset types.
        self.graph.reconfigure_tasks(b=self.b_config)
        self.assertTrue(self.graph.dataset_types.is_resolved("input_1"))
        self.assertFalse(self.graph.dataset_types.is_resolved("output_1"))
        self.assertFalse(self.graph.dataset_types.is_resolved("schema"))
        self.assertFalse(self.graph.dataset_types.is_resolved("intermediate_1"))
        # Resolving again will pick up the new storage class
        self.graph.resolve(MockRegistry(self.dimensions, {}))
        self.assertEqual(
            self.graph.dataset_types["output_1"].storage_class_name, get_mock_name("TaskMetadata")
        )

    def check_visualization(self, graph: PipelineGraph, expected: str, **kwargs: Any) -> None:
        """Run pipeline graph visualization with the given kwargs and check
        that the output is the given expected string.

        Parameters
        ----------
        graph : `lsst.pipe.base.pipeline_graph.PipelineGraph`
            Pipeline graph to visualize.
        expected : `str`
            Expected output of the visualization.  Will be passed through
            `textwrap.dedent`, to allow it to be written with triple-quotes.
        **kwargs
            Forwarded to `lsst.pipe.base.pipeline_graph.visualization.show`.
        """
        stream = io.StringIO()
        visualization.show(graph, stream, **kwargs)
        self.assertEqual(textwrap.dedent(expected), stream.getvalue())

    def test_unresolved_visualization(self) -> None:
        """Test pipeline graph text-based visualization on unresolved
        graphs.
        """
        self.check_visualization(
            self.graph,
            """
            ■  a
            │
            ■  b
            """,
            merge_input_trees=0,
            merge_output_trees=0,
            merge_intermediates=False,
        )
        self.check_visualization(
            self.graph,
            """
            ○  input_1
            │
            ■  a
            │
            ○  intermediate_1
            │
            ■  b
            │
            ○  output_1
            """,
            dataset_types=True,
        )

    def test_resolved_visualization(self) -> None:
        """Test pipeline graph text-based visualization on resolved graphs."""
        self.graph.resolve(MockRegistry(dimensions=self.dimensions, dataset_types={}))
        self.check_visualization(
            self.graph,
            """
            ■  a: {} DynamicTestPipelineTask
            │
            ■  b: {} DynamicTestPipelineTask
            """,
            task_classes="concise",
            merge_input_trees=0,
            merge_output_trees=0,
            merge_intermediates=False,
        )
        self.check_visualization(
            self.graph,
            """
            ○  input_1: {} _mock_StructuredDataDict
            │
            ■  a: {} lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
            │
            ○  intermediate_1: {} _mock_StructuredDataDict
            │
            ■  b: {} lsst.pipe.base.tests.mocks.DynamicTestPipelineTask
            │
            ○  output_1: {} _mock_StructuredDataDict
            """,
            task_classes="full",
            dataset_types=True,
        )


def _have_example_storage_classes() -> bool:
    """Check whether some storage classes work as expected.

    Given that these have registered converters, it shouldn't actually be
    necessary to import be able to those types in order to determine that
    they're convertible, but the storage class machinery is implemented such
    that types that can't be imported can't be converted, and while that's
    inconvenient here it's totally fine in non-testing scenarios where you only
    care about a storage class if you can actually use it.
    """
    getter = StorageClassFactory().getStorageClass
    return (
        getter("ArrowTable").can_convert(getter("ArrowAstropy"))
        and getter("ArrowAstropy").can_convert(getter("ArrowTable"))
        and getter("ArrowTable").can_convert(getter("DataFrame"))
        and getter("DataFrame").can_convert(getter("ArrowTable"))
    )


class PipelineGraphResolveTestCase(unittest.TestCase):
    """More extensive tests for PipelineGraph.resolve and its primate helper
    methods.

    These are in a separate TestCase because they utilize a different `setUp`
    from the rest of the `PipelineGraph` tests.
    """

    def setUp(self) -> None:
        self.a_config = DynamicTestPipelineTaskConfig()
        self.b_config = DynamicTestPipelineTaskConfig()
        self.dimensions = DimensionUniverse()
        self.maxDiff = None

    def make_graph(self) -> PipelineGraph:
        graph = PipelineGraph()
        graph.add_task("a", DynamicTestPipelineTask, self.a_config)
        graph.add_task("b", DynamicTestPipelineTask, self.b_config)
        return graph

    def test_prerequisite_inconsistency(self) -> None:
        """Test that we raise an exception when one edge defines a dataset type
        as a prerequisite and another does not.

        This test will hopefully someday go away (along with
        `DatasetTypeNode.is_prerequisite`) when the QuantumGraph generation
        algorithm becomes more flexible.
        """
        self.a_config.prerequisite_inputs["p"] = DynamicConnectionConfig(dataset_type_name="d")
        self.b_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d")
        graph = self.make_graph()
        with self.assertRaises(ConnectionTypeConsistencyError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_prerequisite_inconsistency_reversed(self) -> None:
        """Same as `test_prerequisite_inconsistency`, with the order the edges
        are added to the graph reversed.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d")
        self.b_config.prerequisite_inputs["p"] = DynamicConnectionConfig(dataset_type_name="d")
        graph = self.make_graph()
        with self.assertRaises(ConnectionTypeConsistencyError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_prerequisite_output(self) -> None:
        """Test that we raise an exception when one edge defines a dataset type
        as a prerequisite but another defines it as an output.
        """
        self.a_config.prerequisite_inputs["p"] = DynamicConnectionConfig(dataset_type_name="d")
        self.b_config.outputs["o"] = DynamicConnectionConfig(dataset_type_name="d")
        graph = self.make_graph()
        with self.assertRaises(ConnectionTypeConsistencyError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_skypix_missing(self) -> None:
        """Test that we raise an exception when one edge uses the "skypix"
        dimension as a placeholder but the dataset type is not registered.
        """
        self.a_config.prerequisite_inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", dimensions={"skypix"}
        )
        graph = self.make_graph()
        with self.assertRaises(MissingDatasetTypeError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_skypix_inconsistent(self) -> None:
        """Test that we raise an exception when one edge uses the "skypix"
        dimension as a placeholder but the rest of the dimensions are
        inconsistent with the registered dataset type.
        """
        self.a_config.prerequisite_inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", dimensions={"skypix", "visit"}
        )
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {
                        "d": DatasetType(
                            "d",
                            dimensions=self.dimensions.conform(["htm7"]),
                            storageClass="StructuredDataDict",
                        )
                    },
                )
            )
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {
                        "d": DatasetType(
                            "d",
                            dimensions=self.dimensions.conform(["htm7", "visit", "skymap"]),
                            storageClass="StructuredDataDict",
                        )
                    },
                )
            )

    def test_duplicate_outputs(self) -> None:
        """Test that we raise an exception when a dataset type node would have
        two write edges.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(dataset_type_name="d")
        self.b_config.outputs["o"] = DynamicConnectionConfig(dataset_type_name="d")
        graph = self.make_graph()
        with self.assertRaises(DuplicateOutputError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_component_of_unregistered_parent(self) -> None:
        """Test that we raise an exception when a component dataset type's
        parent is not registered.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d.c")
        graph = self.make_graph()
        with self.assertRaises(MissingDatasetTypeError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_undefined_component(self) -> None:
        """Test that we raise an exception when a component dataset type's
        parent is registered, but its storage class does not have that
        component.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d.c")
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {"d": DatasetType("d", self.dimensions.empty, get_mock_name("StructuredDataDict"))},
                )
            )

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_bad_component_storage_class(self) -> None:
        """Test that we raise an exception when a component dataset type's
        parent is registered, but does not have that component.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d.schema", storage_class="StructuredDataDict"
        )
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {"d": DatasetType("d", self.dimensions.empty, get_mock_name("ArrowTable"))},
                )
            )

    def test_input_storage_class_incompatible_with_registry(self) -> None:
        """Test that we raise an exception when an input connection's storage
        class is incompatible with the registry definition.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataList"
        )
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {"d": DatasetType("d", self.dimensions.empty, get_mock_name("StructuredDataDict"))},
                )
            )

    def test_output_storage_class_incompatible_with_registry(self) -> None:
        """Test that we raise an exception when an output connection's storage
        class is incompatible with the registry definition.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataList"
        )
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(
                MockRegistry(
                    self.dimensions,
                    {"d": DatasetType("d", self.dimensions.empty, get_mock_name("StructuredDataDict"))},
                )
            )

    def test_input_storage_class_incompatible_with_output(self) -> None:
        """Test that we raise an exception when an input connection's storage
        class is incompatible with the storage class of the output connection.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataDict"
        )
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataList"
        )
        graph = self.make_graph()
        with self.assertRaises(IncompatibleDatasetTypeError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    def test_ambiguous_storage_class(self) -> None:
        """Test that we raise an exception when two input connections define
        the same dataset with different storage classes (even compatible ones)
        and there is no output connection or registry definition to take
        precedence.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataDict"
        )
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="StructuredDataList"
        )
        graph = self.make_graph()
        with self.assertRaises(MissingDatasetTypeError):
            graph.resolve(MockRegistry(self.dimensions, {}))

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_inputs_compatible_with_registry(self) -> None:
        """Test successful resolution of a dataset type where input edges have
        different but compatible storage classes and the dataset type is
        already registered.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d", storage_class="ArrowTable")
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="ArrowAstropy"
        )
        graph = self.make_graph()
        dataset_type = DatasetType("d", self.dimensions.empty, get_mock_name("DataFrame"))
        graph.resolve(MockRegistry(self.dimensions, {"d": dataset_type}))
        self.assertEqual(graph.dataset_types["d"].dataset_type, dataset_type)
        a_i = graph.tasks["a"].inputs["i"]
        b_i = graph.tasks["b"].inputs["i"]
        self.assertEqual(
            a_i.adapt_dataset_type(dataset_type),
            dataset_type.overrideStorageClass(get_mock_name("ArrowTable")),
        )
        self.assertEqual(
            b_i.adapt_dataset_type(dataset_type),
            dataset_type.overrideStorageClass(get_mock_name("ArrowAstropy")),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(dataset_type, data_id, run="r")
        a_ref = a_i.adapt_dataset_ref(ref)
        b_ref = b_i.adapt_dataset_ref(ref)
        self.assertEqual(a_ref, ref.overrideStorageClass(get_mock_name("ArrowTable")))
        self.assertEqual(b_ref, ref.overrideStorageClass(get_mock_name("ArrowAstropy")))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(a_ref), ref)
        self.assertEqual(graph.dataset_types["d"].generalize_ref(b_ref), ref)

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_output_compatible_with_registry(self) -> None:
        """Test successful resolution of a dataset type where an output edge
        has a different but compatible storage class from the dataset type
        already registered.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="ArrowTable"
        )
        graph = self.make_graph()
        dataset_type = DatasetType("d", self.dimensions.empty, get_mock_name("DataFrame"))
        graph.resolve(MockRegistry(self.dimensions, {"d": dataset_type}))
        self.assertEqual(graph.dataset_types["d"].dataset_type, dataset_type)
        a_o = graph.tasks["a"].outputs["o"]
        self.assertEqual(
            a_o.adapt_dataset_type(dataset_type),
            dataset_type.overrideStorageClass(get_mock_name("ArrowTable")),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(dataset_type, data_id, run="r")
        a_ref = a_o.adapt_dataset_ref(ref)
        self.assertEqual(a_ref, ref.overrideStorageClass(get_mock_name("ArrowTable")))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(a_ref), ref)

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_inputs_compatible_with_output(self) -> None:
        """Test successful resolution of a dataset type where an input edge has
        a different but compatible storage class from the output edge, and
        the dataset type is not registered.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="ArrowTable"
        )
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="ArrowAstropy"
        )
        graph = self.make_graph()
        a_o = graph.tasks["a"].outputs["o"]
        b_i = graph.tasks["b"].inputs["i"]
        graph.resolve(MockRegistry(self.dimensions, {}))
        self.assertEqual(graph.dataset_types["d"].storage_class_name, get_mock_name("ArrowTable"))
        self.assertEqual(
            a_o.adapt_dataset_type(graph.dataset_types["d"].dataset_type),
            graph.dataset_types["d"].dataset_type,
        )
        self.assertEqual(
            b_i.adapt_dataset_type(graph.dataset_types["d"].dataset_type),
            graph.dataset_types["d"].dataset_type.overrideStorageClass(get_mock_name("ArrowAstropy")),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(graph.dataset_types["d"].dataset_type, data_id, run="r")
        a_ref = a_o.adapt_dataset_ref(ref)
        b_ref = b_i.adapt_dataset_ref(ref)
        self.assertEqual(a_ref, ref)
        self.assertEqual(b_ref, ref.overrideStorageClass(get_mock_name("ArrowAstropy")))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(a_ref), ref)
        self.assertEqual(graph.dataset_types["d"].generalize_ref(b_ref), ref)

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_component_resolved_by_input(self) -> None:
        """Test successful resolution of a component dataset type due to
        another input referencing the parent dataset type.
        """
        self.a_config.inputs["i"] = DynamicConnectionConfig(dataset_type_name="d", storage_class="ArrowTable")
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d.schema", storage_class="ArrowSchema"
        )
        graph = self.make_graph()
        parent_dataset_type = DatasetType("d", self.dimensions.empty, get_mock_name("ArrowTable"))
        graph.resolve(MockRegistry(self.dimensions, {}))
        self.assertEqual(graph.dataset_types["d"].dataset_type, parent_dataset_type)
        a_i = graph.tasks["a"].inputs["i"]
        b_i = graph.tasks["b"].inputs["i"]
        self.assertEqual(b_i.dataset_type_name, "d.schema")
        self.assertEqual(a_i.adapt_dataset_type(parent_dataset_type), parent_dataset_type)
        self.assertEqual(
            b_i.adapt_dataset_type(parent_dataset_type),
            parent_dataset_type.makeComponentDatasetType("schema"),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(parent_dataset_type, data_id, run="r")
        a_ref = a_i.adapt_dataset_ref(ref)
        b_ref = b_i.adapt_dataset_ref(ref)
        self.assertEqual(a_ref, ref)
        self.assertEqual(b_ref, ref.makeComponentRef("schema"))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(a_ref), ref)
        self.assertEqual(graph.dataset_types["d"].generalize_ref(b_ref), ref)

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_component_resolved_by_output(self) -> None:
        """Test successful resolution of a component dataset type due to
        an output connection referencing the parent dataset type.
        """
        self.a_config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="d", storage_class="ArrowTable"
        )
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d.schema", storage_class="ArrowSchema"
        )
        graph = self.make_graph()
        parent_dataset_type = DatasetType("d", self.dimensions.empty, get_mock_name("ArrowTable"))
        graph.resolve(MockRegistry(self.dimensions, {}))
        self.assertEqual(graph.dataset_types["d"].dataset_type, parent_dataset_type)
        a_o = graph.tasks["a"].outputs["o"]
        b_i = graph.tasks["b"].inputs["i"]
        self.assertEqual(b_i.dataset_type_name, "d.schema")
        self.assertEqual(a_o.adapt_dataset_type(parent_dataset_type), parent_dataset_type)
        self.assertEqual(
            b_i.adapt_dataset_type(parent_dataset_type),
            parent_dataset_type.makeComponentDatasetType("schema"),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(parent_dataset_type, data_id, run="r")
        a_ref = a_o.adapt_dataset_ref(ref)
        b_ref = b_i.adapt_dataset_ref(ref)
        self.assertEqual(a_ref, ref)
        self.assertEqual(b_ref, ref.makeComponentRef("schema"))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(a_ref), ref)
        self.assertEqual(graph.dataset_types["d"].generalize_ref(b_ref), ref)

    @unittest.skipUnless(
        _have_example_storage_classes(), "Arrow/Astropy/Pandas storage classes are not available."
    )
    def test_component_resolved_by_registry(self) -> None:
        """Test successful resolution of a component dataset type due to
        the parent dataset type already being registered.
        """
        self.b_config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="d.schema", storage_class="ArrowSchema"
        )
        graph = self.make_graph()
        parent_dataset_type = DatasetType("d", self.dimensions.empty, get_mock_name("ArrowTable"))
        graph.resolve(MockRegistry(self.dimensions, {"d": parent_dataset_type}))
        self.assertEqual(graph.dataset_types["d"].dataset_type, parent_dataset_type)
        b_i = graph.tasks["b"].inputs["i"]
        self.assertEqual(b_i.dataset_type_name, "d.schema")
        self.assertEqual(
            b_i.adapt_dataset_type(parent_dataset_type),
            parent_dataset_type.makeComponentDatasetType("schema"),
        )
        data_id = DataCoordinate.make_empty(self.dimensions)
        ref = DatasetRef(parent_dataset_type, data_id, run="r")
        b_ref = b_i.adapt_dataset_ref(ref)
        self.assertEqual(b_ref, ref.makeComponentRef("schema"))
        self.assertEqual(graph.dataset_types["d"].generalize_ref(b_ref), ref)


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



#
# LSST Data Management System
# Copyright 2008, 2009, 2010, 2011 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

from __future__ import annotations

__all__ = ["Struct"]

from types import SimpleNamespace
from typing import Any


class Struct(SimpleNamespace):
    """A container to which you can add fields as attributes.

    Parameters
    ----------
    **keyArgs
        Keyword arguments specifying fields and their values.

    Notes
    -----
    Intended to be used for the return value from `~lsst.pipe.base.Task.run`
    and other `~lsst.pipe.base.Task` methods, and useful for any method that
    returns multiple values.

    The intent is to allow accessing returned items by name, instead of
    unpacking a tuple.  This makes the code much more robust and easier to
    read. It allows one to change what values are returned without inducing
    mysterious failures: adding items is completely safe, and removing or
    renaming items causes errors that are caught quickly and reported in a way
    that is easy to understand.

    The primary reason for using Struct instead of dict is that the fields may
    be accessed as attributes, e.g. ``aStruct.foo`` instead of
    ``aDict["foo"]``. Admittedly this only saves a few characters, but it
    makes the code significantly more readable.

    Struct is preferred over named tuples, because named tuples can be used as
    ordinary tuples, thus losing all the safety advantages of Struct. In
    addition, named tuples are clumsy to define and Structs are much more
    mutable (e.g. one can trivially combine Structs and add additional fields).

    Examples
    --------
    >>> myStruct = Struct(
    >>>     strVal = 'the value of the field named "strVal"',
    >>>     intVal = 35,
    >>> )
    """

    def __init__(self, **keyArgs: Any):
        for name, val in keyArgs.items():
            self.__safeAdd(name, val)

    def __safeAdd(self, name: str, val: Any) -> None:
        """Add a field if it does not already exist and name does not start
        with ``__`` (two underscores).

        Parameters
        ----------
        name : `str`
            Name of field to add.
        val : object
            Value of field to add.

        Raises
        ------
        RuntimeError
            Raised if name already exists or starts with ``__`` (two
            underscores).
        """
        if hasattr(self, name):
            raise RuntimeError(f"Item {name!r} already exists")
        if name.startswith("__"):
            raise RuntimeError(f"Item name {name!r} invalid; must not begin with __")
        setattr(self, name, val)

    def getDict(self) -> dict[str, Any]:
        """Get a dictionary of fields in this struct.

        Returns
        -------
        structDict : `dict`
            Dictionary with field names as keys and field values as values.
            The values are shallow copies.
        """
        return self.__dict__.copy()

    def mergeItems(self, struct: Struct, *nameList: str) -> None:
        """Copy specified fields from another struct, provided they don't
        already exist.

        Parameters
        ----------
        struct : `Struct`
            `Struct` from which to copy.
        *nameList : `str`
            All remaining arguments are names of items to copy.

        Raises
        ------
        RuntimeError
            Raised if any item in nameList already exists in self (but any
            items before the conflicting item in nameList will have been
            copied).

        Examples
        --------
        For example::

            foo.copyItems(other, "itemName1", "itemName2")

        copies ``other.itemName1`` and ``other.itemName2`` into self.
        """
        for name in nameList:
            self.__safeAdd(name, getattr(struct, name))

    def copy(self) -> Struct:
        """Make a one-level-deep copy (values are not copied).

        Returns
        -------
        copy : `Struct`
            One-level-deep copy of this Struct.
        """
        return Struct(**self.getDict())

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        itemsStr = "; ".join(f"{name}={val}" for name, val in self.getDict().items())
        return f"{self.__class__.__name__}({itemsStr})"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "NoWorkFound",
    "RepeatableQuantumError",
    "InvalidQuantumError",
)


class NoWorkFound(BaseException):
    """An exception raised when a Quantum should not exist because there is no
    work for it to do.

    This usually occurs because a non-optional input dataset is not present, or
    a spatiotemporal overlap that was conservatively predicted does not
    actually exist.

    This inherits from BaseException because it is used to signal a case that
    we don't consider a real error, even though we often want to use try/except
    logic to trap it.
    """


class RepeatableQuantumError(RuntimeError):
    """Exception that may be raised by PipelineTasks (and code they delegate
    to) in order to indicate that a repeatable problem that will not be
    addressed by retries.

    This usually indicates that the algorithm and the data it has been given
    are somehow incompatible, and the task should run fine on most other data.

    This exception may be used as a base class for more specific questions, or
    used directly while chaining another exception, e.g.::

        try:
            run_code()
        except SomeOtherError as err:
            raise RepeatableQuantumError() from err

    This may be used for missing input data when the desired behavior is to
    cause all downstream tasks being run be blocked, forcing the user to
    address the problem.  When the desired behavior is to skip all of this
    quantum and attempt downstream tasks (or skip them) without its its
    outputs, raise `NoWorkFound` or return without raising instead.
    """

    EXIT_CODE = 20


class InvalidQuantumError(Exception):
    """Exception that may be raised by PipelineTasks (and code they delegate
    to) in order to indicate logic bug or configuration problem.

    This usually indicates that the configured algorithm itself is invalid and
    will not run on a significant fraction of quanta (often all of them).

    This exception may be used as a base class for more specific questions, or
    used directly while chaining another exception, e.g.::

        try:
            run_code()
        except SomeOtherError as err:
            raise RepeatableQuantumError() from err

    Raising this exception in `PipelineTask.runQuantum` or something it calls
    is a last resort - whenever possible, such problems should cause exceptions
    in ``__init__`` or in QuantumGraph generation.  It should never be used
    for missing data.
    """

    EXIT_CODE = 21



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The base class for the QuantumGraph-generation algorithm and various
helper classes.
"""

from __future__ import annotations

__all__ = (
    "QuantumGraphBuilder",
    "ExistingDatasets",
    "QuantumGraphBuilderError",
    "OutputExistsError",
    "PrerequisiteMissingError",
)

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, final

from lsst.daf.butler import (
    Butler,
    CollectionType,
    DataCoordinate,
    DatasetRef,
    DatasetType,
    DimensionUniverse,
    NamedKeyDict,
    NamedKeyMapping,
    Quantum,
)
from lsst.daf.butler.registry import MissingCollectionError, MissingDatasetTypeError
from lsst.utils.logging import LsstLogAdapter, getLogger
from lsst.utils.timer import timeMethod

from . import automatic_connection_constants as acc
from ._status import NoWorkFound
from ._task_metadata import TaskMetadata
from .connections import AdjustQuantumHelper
from .graph import QuantumGraph
from .pipeline_graph import PipelineGraph, TaskNode
from .prerequisite_helpers import PrerequisiteInfo, SkyPixBoundsBuilder, TimespanBuilder
from .quantum_graph_skeleton import (
    DatasetKey,
    PrerequisiteDatasetKey,
    QuantumGraphSkeleton,
    QuantumKey,
    TaskInitKey,
)

if TYPE_CHECKING:
    from .pipeline import TaskDef


class QuantumGraphBuilderError(Exception):
    """Base class for exceptions generated by QuantumGraphBuilder."""

    pass


class GraphBuilderError(QuantumGraphBuilderError):
    """Backwards-compatibility near-alias for QuantumGraphBuilderError."""

    pass


# Inherit from backwards-compatibility alias for backwards-compatibility.
class OutputExistsError(GraphBuilderError):
    """Exception generated when output datasets already exist."""

    pass


# Inherit from backwards-compatibility alias for backwards-compatibility.
class PrerequisiteMissingError(GraphBuilderError):
    """Exception generated when a prerequisite dataset does not exist."""

    pass


class InitInputMissingError(QuantumGraphBuilderError):
    """Exception generated when an init-input dataset does not exist."""

    pass


class QuantumGraphBuilder(ABC):
    """An abstract base class for building `QuantumGraph` objects from a
    pipeline.

    Parameters
    ----------
    pipeline_graph : `.pipeline_graph.PipelineGraph`
        Pipeline to build a `QuantumGraph` from, as a graph.  Will be resolved
        in-place with the given butler (any existing resolution is ignored).
    butler : `lsst.daf.butler.Butler`
        Client for the data repository.  Should be read-only.
    input_collections : `~collections.abc.Sequence` [ `str` ], optional
        Collections to search for overall-input datasets.  If not provided,
        ``butler.collections`` is used (and must not be empty).
    output_run : `str`, optional
        Output `~lsst.daf.butler.CollectionType.RUN` collection.  If not
        provided, ``butler.run`` is used (and must not be `None`).
    skip_existing_in : `~collections.abc.Sequence` [ `str` ], optional
        Collections to search for outputs that already exist for the purpose of
        skipping quanta that have already been run.
    clobber : `bool`, optional
        Whether to raise if predicted outputs already exist in ``output_run``
        (not including those quanta that would be skipped because they've
        already been run).  This never actually clobbers outputs; it just
        informs the graph generation algorithm whether execution will run with
        clobbering enabled.  This is ignored if ``output_run`` does not exist.

    Notes
    -----
    Constructing a `QuantumGraphBuilder` will run queries for existing datasets
    with empty data IDs (including but not limited to init inputs and outputs),
    in addition to resolving the given pipeline graph and testing for existence
    of the ``output`` run collection.

    The `build` method splits the pipeline graph into independent subgraphs,
    then calls the abstract method `process_subgraph` on each, to allow
    concrete implementations to populate the rough graph structure (the
    `~quantum_graph_skeleton.QuantumGraphSkeleton` class) and search for
    existing datasets (further populating the builder's `existing_datasets`
    struct).  The `build` method then:

    - assembles `lsst.daf.butler.Quantum` instances from all data IDs in the
      skeleton;
    - looks for existing outputs found in ``skip_existing_in`` to see if any
      quanta should be skipped;
    - calls `PipelineTaskConnections.adjustQuantum` on all quanta, adjusting
      downstream quanta appropriately when preliminary predicted outputs are
      rejected (pruning nodes that will not have the inputs they need to run);
    - attaches datastore records and registry dataset types to the graph.

    In addition to implementing `process_subgraph`, derived classes are
    generally expected to add new construction keyword-only arguments to
    control the data IDs of the quantum graph, while forwarding all of the
    arguments defined in the base class to `super`.
    """

    def __init__(
        self,
        pipeline_graph: PipelineGraph,
        butler: Butler,
        *,
        input_collections: Sequence[str] | None = None,
        output_run: str | None = None,
        skip_existing_in: Sequence[str] = (),
        clobber: bool = False,
    ):
        self.log = getLogger(__name__)
        self.metadata = TaskMetadata()
        self._pipeline_graph = pipeline_graph
        self.butler = butler
        if input_collections is None:
            input_collections = butler.collections
        if not input_collections:
            raise ValueError("No input collections provided.")
        self.input_collections = input_collections
        if output_run is None:
            output_run = butler.run
        if not output_run:
            raise ValueError("No output RUN collection provided.")
        self.output_run = output_run
        self.skip_existing_in = skip_existing_in
        self.empty_data_id = DataCoordinate.make_empty(butler.dimensions)
        self.clobber = clobber
        # See whether the output run already exists.
        self.output_run_exists = False
        try:
            if self.butler.registry.getCollectionType(self.output_run) is not CollectionType.RUN:
                raise RuntimeError(f"{self.output_run!r} is not a RUN collection.")
            self.output_run_exists = True
        except MissingCollectionError:
            # If the run doesn't exist we never need to clobber.  This is not
            # an error so you can run with clobber=True the first time you
            # attempt some processing as well as all subsequent times, instead
            # of forcing the user to make the first attempt different.
            self.clobber = False
        # We need to know whether the skip_existing_in collection sequence
        # starts with the output run collection, as an optimization to avoid
        # queries later.
        try:
            skip_existing_in_flat = self.butler.registry.queryCollections(
                self.skip_existing_in, flattenChains=True
            )
        except MissingCollectionError:
            skip_existing_in_flat = []
        if not skip_existing_in_flat:
            self.skip_existing_in = []
        if self.skip_existing_in and self.output_run_exists:
            self.skip_existing_starts_with_output_run = self.output_run == skip_existing_in_flat[0]
        else:
            self.skip_existing_starts_with_output_run = False
        self.existing_datasets = ExistingDatasets()
        try:
            packages_storage_class = butler.get_dataset_type(acc.PACKAGES_INIT_OUTPUT_NAME).storageClass_name
        except MissingDatasetTypeError:
            packages_storage_class = acc.PACKAGES_INIT_OUTPUT_STORAGE_CLASS
        self._global_init_output_types = {
            acc.PACKAGES_INIT_OUTPUT_NAME: DatasetType(
                acc.PACKAGES_INIT_OUTPUT_NAME,
                self.universe.empty,
                packages_storage_class,
            )
        }
        with self.butler.registry.caching_context():
            self._pipeline_graph.resolve(self.butler.registry)
            self._find_empty_dimension_datasets()
            self.prerequisite_info = {
                task_node.label: PrerequisiteInfo(task_node, self._pipeline_graph)
                for task_node in pipeline_graph.tasks.values()
            }

    log: LsstLogAdapter
    """Logger to use for all quantum-graph generation messages.

    General and per-task status messages should be logged at `~logging.INFO`
    level or higher, per-dataset-type status messages should be logged at
    `~lsst.utils.logging.VERBOSE` or higher, and per-data-ID status messages
    should be logged at `logging.DEBUG` or higher.
    """

    metadata: TaskMetadata
    """Metadata to store in the QuantumGraph.

    The `TaskMetadata` class is used here primarily in order to enable
    resource-usage collection with the `lsst.utils.timer.timeMethod` decorator.
    """

    butler: Butler
    """Client for the data repository.

    Should be read-only.
    """

    input_collections: Sequence[str]
    """Collections to search for overall-input datasets.
    """

    output_run: str
    """Output `~lsst.daf.butler.CollectionType.RUN` collection.
    """

    skip_existing_in: Sequence[str]
    """Collections to search for outputs that already exist for the purpose
    of skipping quanta that have already been run.
    """

    clobber: bool
    """Whether to raise if predicted outputs already exist in ``output_run``

    This never actually clobbers outputs; it just informs the graph generation
    algorithm whether execution will run with clobbering enabled.  This is
    always `False` if `output_run_exists` is `False`.
    """

    empty_data_id: DataCoordinate
    """An empty data ID in the data repository's dimension universe.
    """

    output_run_exists: bool
    """Whether the output run exists in the data repository already.
    """

    skip_existing_starts_with_output_run: bool
    """Whether the `skip_existing_in` sequence begins with `output_run`.

    If this is true, any dataset found in `output_run` can be used to
    short-circuit queries in `skip_existing_in`.
    """

    existing_datasets: ExistingDatasets
    """Struct holding datasets that have already been found in the data
    repository.

    This is updated in-place as the `QuantumGraph` generation algorithm
    proceeds.
    """

    prerequisite_info: Mapping[str, PrerequisiteInfo]
    """Helper objects for finding prerequisite inputs, organized by task label.

    Subclasses that find prerequisites should remove the
    covered `~prerequisite_helpers.PrerequisiteFinder` objects from this
    attribute.
    """

    @property
    def universe(self) -> DimensionUniverse:
        """Definitions of all data dimensions."""
        return self.butler.dimensions

    @final
    @timeMethod
    def build(self, metadata: Mapping[str, Any] | None = None) -> QuantumGraph:
        """Build the quantum graph.

        Parameters
        ----------
        metadata : `~collections.abc.Mapping`, optional
            Flexible metadata to add to the quantum graph.

        Returns
        -------
        quantum_graph : `QuantumGraph`
            DAG describing processing to be performed.

        Notes
        -----
        External code is expected to construct a `QuantumGraphBuilder` and then
        call this method exactly once.  See class documentation for details on
        what it does.
        """
        with self.butler.registry.caching_context():
            full_skeleton = QuantumGraphSkeleton(self._pipeline_graph.tasks)
            subgraphs = list(self._pipeline_graph.split_independent())
            for i, subgraph in enumerate(subgraphs):
                self.log.info(
                    "Processing pipeline subgraph %d of %d with %d task(s).",
                    i + 1,
                    len(subgraphs),
                    len(subgraph.tasks),
                )
                self.log.verbose("Subgraph tasks: [%s]", ", ".join(label for label in subgraph.tasks))
                subgraph_skeleton = self.process_subgraph(subgraph)
                full_skeleton.update(subgraph_skeleton)
            # Loop over tasks.  The pipeline graph must be topologically
            # sorted, so a quantum is only processed after any quantum that
            # provides its inputs has been processed.
            for task_node in self._pipeline_graph.tasks.values():
                self._resolve_task_quanta(task_node, full_skeleton)
            # Add global init-outputs to the skeleton.
            for dataset_type in self._global_init_output_types.values():
                dataset_key = full_skeleton.add_dataset_node(
                    dataset_type.name, self.empty_data_id, is_global_init_output=True
                )
                ref = self.existing_datasets.outputs_in_the_way.get(dataset_key)
                if ref is None:
                    ref = DatasetRef(dataset_type, self.empty_data_id, run=self.output_run)
                full_skeleton[dataset_key]["ref"] = ref
            # Remove dataset nodes with no edges that are not global init
            # outputs, which are generally overall-inputs whose original quanta
            # end up skipped or with no work to do (we can't remove these along
            # with the quanta because no quantum knows if its the only
            # consumer).
            full_skeleton.remove_orphan_datasets()
            self._attach_datastore_records(full_skeleton)
            # TODO initialize most metadata here instead of in ctrl_mpexec.
            if metadata is None:
                metadata = {}
            return self._construct_quantum_graph(full_skeleton, metadata)

    @abstractmethod
    def process_subgraph(self, subgraph: PipelineGraph) -> QuantumGraphSkeleton:
        """Build the rough structure for an independent subset of the
        `QuantumGraph` and query for relevant existing datasets.

        Parameters
        ----------
        subgraph : `.pipeline_graph.PipelineGraph`
            Subset of the pipeline graph that should be processed by this call.
            This is always resolved and topologically sorted.  It should not be
            modified.

        Returns
        -------
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Class representing an initial quantum graph. See
            `quantum_graph_skeleton.QuantumGraphSkeleton` docs for details.
            After this is returned, the object may be modified in-place in
            unspecified ways.

        Notes
        -----
        In addition to returning a
        `quantum_graph_skeleton.QuantumGraphSkeleton`, this method should
        populate the `existing_datasets` structure by querying for all relevant
        datasets with non-empty data IDs (those with empty data IDs will
        already be present).  In particular:

        - `~ExistingDatasets.inputs` must always be populated with all
          overall-input datasets (but not prerequisites), by querying
          `input_collections`;
        - `~ExistingDatasets.outputs_for_skip` must be populated with any
          intermediate our output datasets present in `skip_existing_in` (it
          can be ignored if `skip_existing_in` is empty);
        - `~ExistingDatasets.outputs_in_the_way` must be populated with any
          intermediate or output datasets present in `output_run`, if
          `output_run_exists` (it can be ignored if `output_run_exists` is
          `False`).  Note that the presence of such datasets is not
          automatically an error, even if `clobber is `False`, as these may be
          quanta that will be skipped.
        - `~ExistingDatasets.inputs` must be populated with all
          prerequisite-input datasets that were included in the skeleton, by
          querying `input_collections` (not all prerequisite inputs need to be
          included in the skeleton, but the base class can only use per-quantum
          queries to find them, and that can be slow when there are many
          quanta).

        Dataset types should never be components and should always use the
        "common" storage class definition in `pipeline_graph.DatasetTypeNode`
        (which is the data repository definition when the dataset type is
        registered).
        """
        raise NotImplementedError()

    @final
    @timeMethod
    def _resolve_task_quanta(self, task_node: TaskNode, skeleton: QuantumGraphSkeleton) -> None:
        """Process the quanta for one task in a skeleton graph to skip those
        that have already completed and adjust those that request it.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Notes
        -----
        This method modifies ``skeleton`` in-place in several ways:

        - It adds a "ref" attribute to dataset nodes, using the contents of
          `existing_datasets`.  This ensures producing and consuming tasks
          start from the same `DatasetRef`.
        - It adds "inputs", "outputs", and "init_inputs" attributes to the
          quantum nodes, holding the same `NamedValueMapping` objects needed to
          construct an actual `Quantum` instances.
        - It removes quantum nodes that are to be skipped because their outputs
          already exist in `skip_existing_in`.  It also removes their outputs
          from `ExistingDatasets.outputs_in_the_way`.
        - It adds prerequisite dataset nodes and edges that connect them to the
          quanta that consume them.
        - It removes quantum nodes whose
          `~PipelineTaskConnections.adjustQuantum` calls raise `NoWorkFound` or
          predict no outputs;
        - It removes the nodes of output datasets that are "adjusted away".
        - It removes the edges of input datasets that are "adjusted away".

        The difference between how adjusted inputs and outputs are handled
        reflects the fact that many quanta can share the same input, but only
        one produces each output.  This can lead to the graph having
        superfluous isolated nodes after processing is complete, but these
        should only be removed after all the quanta from all tasks have been
        processed.
        """
        # Extract the helper object for the prerequisite inputs of this task,
        # and tell it to prepare to construct skypix bounds and timespans for
        # each quantum (these will automatically do nothing if nothing needs
        # those bounds).
        task_prerequisite_info = self.prerequisite_info[task_node.label]
        task_prerequisite_info.update_bounds()
        # Loop over all quanta for this task, remembering the ones we've
        # gotten rid of.
        skipped_quanta = []
        no_work_quanta = []
        for quantum_key in skeleton.get_quanta(task_node.label):
            if self._skip_quantum_if_metadata_exists(task_node, quantum_key, skeleton):
                skipped_quanta.append(quantum_key)
                continue
            quantum_data_id = skeleton[quantum_key]["data_id"]
            skypix_bounds_builder = task_prerequisite_info.bounds.make_skypix_bounds_builder(quantum_data_id)
            timespan_builder = task_prerequisite_info.bounds.make_timespan_builder(quantum_data_id)
            adjusted_outputs = self._gather_quantum_outputs(
                task_node, quantum_key, skeleton, skypix_bounds_builder, timespan_builder
            )
            adjusted_inputs = self._gather_quantum_inputs(
                task_node,
                quantum_key,
                skeleton,
                task_prerequisite_info,
                skypix_bounds_builder,
                timespan_builder,
            )
            # Give the task's Connections class an opportunity to remove
            # some inputs, or complain if they are unacceptable.  This will
            # raise if one of the check conditions is not met, which is the
            # intended behavior.
            helper = AdjustQuantumHelper(inputs=adjusted_inputs, outputs=adjusted_outputs)
            try:
                helper.adjust_in_place(
                    task_node._get_imported_data().connections, task_node.label, quantum_data_id
                )
            except NoWorkFound as err:
                # Do not generate this quantum; it would not produce any
                # outputs.  Remove it and all of the outputs it might have
                # produced from the skeleton.
                try:
                    _, connection_name, _ = err.args
                    details = f"not enough datasets for connection {connection_name}."
                except ValueError:
                    details = str(err)
                self.log.debug(
                    "No work found for quantum %s of task %s: %s",
                    quantum_key.data_id_values,
                    quantum_key.task_label,
                    details,
                )
                no_work_quanta.append(quantum_key)
                continue
            if helper.outputs_adjusted:
                if not any(adjusted_refs for adjusted_refs in helper.outputs.values()):
                    # No outputs also means we don't generate this quantum.
                    self.log.debug(
                        "No outputs predicted for quantum %s of task %s.",
                        quantum_key.data_id_values,
                        quantum_key.task_label,
                    )
                    no_work_quanta.append(quantum_key)
                    continue
                # Remove output nodes that were not retained by
                # adjustQuantum.
                skeleton.remove_dataset_nodes(
                    self._find_removed(skeleton.iter_outputs_of(quantum_key), helper.outputs)
                )
            if helper.inputs_adjusted:
                if not any(bool(adjusted_refs) for adjusted_refs in helper.inputs.values()):
                    raise QuantumGraphBuilderError(
                        f"adjustQuantum implementation for {task_node.label}@{quantum_key.data_id_values} "
                        "returned outputs but no inputs."
                    )
                # Remove input dataset edges that were not retained by
                # adjustQuantum.  We can't remove the input dataset nodes
                # because some other quantum might still want them.
                skeleton.remove_input_edges(
                    quantum_key, self._find_removed(skeleton.iter_inputs_of(quantum_key), helper.inputs)
                )
            # Save the adjusted inputs and outputs to the quantum node's
            # state so we don't have to regenerate those data structures
            # from the graph.
            skeleton[quantum_key]["inputs"] = helper.inputs
            skeleton[quantum_key]["outputs"] = helper.outputs
        for no_work_quantum in no_work_quanta:
            skeleton.remove_quantum_node(no_work_quantum, remove_outputs=True)
        for skipped_quantum in skipped_quanta:
            skeleton.remove_quantum_node(skipped_quantum, remove_outputs=False)
        remaining_quanta = skeleton.get_quanta(task_node.label)
        self._resolve_task_init(task_node, skeleton, bool(skipped_quanta))
        message_terms = []
        if no_work_quanta:
            message_terms.append(f"{len(no_work_quanta)} had no work to do")
        if skipped_quanta:
            message_terms.append(f"{len(no_work_quanta)} previously succeeded")
        message_parenthetical = f" ({', '.join(message_terms)})" if message_terms else ""
        if remaining_quanta:
            self.log.info(
                "Generated %s for task %s%s.",
                _quantum_or_quanta(len(remaining_quanta)),
                task_node.label,
                message_parenthetical,
            )
        else:
            self.log.info(
                "Dropping task %s because no quanta remain%s.", task_node.label, message_parenthetical
            )
            skeleton.remove_task(task_node.label)

    def _skip_quantum_if_metadata_exists(
        self, task_node: TaskNode, quantum_key: QuantumKey, skeleton: QuantumGraphSkeleton
    ) -> bool:
        """Identify and drop quanta that should be skipped because their
        metadata datasets already exist.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.

        Returns
        -------
        skipped : `bool`
            `True` if the quantum is being skipped and has been removed from
            the graph, `False` otherwise.

        Notes
        -----
        If the metadata dataset for this quantum exists in
        `ExistingDatasets.outputs_for_skip`, the quantum will be skipped. This
        causes the quantum node to be removed from the graph.  Dataset nodes
        that were previously the outputs of this quantum will have their "ref"
        attribute set from `ExistingDatasets.outputs_for_skip`, or will be
        removed if there is no such dataset there.  Any output dataset in
        `ExistingDatasets.outputs_in_the_way` will be removed.
        """
        metadata_dataset_key = DatasetKey(
            task_node.metadata_output.parent_dataset_type_name, quantum_key.data_id_values
        )
        if metadata_dataset_key in self.existing_datasets.outputs_for_skip:
            # This quantum's metadata is already present in the the
            # skip_existing_in collections; we'll skip it.  But the presence of
            # the metadata dataset doesn't guarantee that all of the other
            # outputs we predicted are present; we have to check.
            for output_dataset_key in list(skeleton.iter_outputs_of(quantum_key)):
                if (
                    output_ref := self.existing_datasets.outputs_for_skip.get(output_dataset_key)
                ) is not None:
                    # Populate the skeleton graph's node attributes
                    # with the existing DatasetRef, just like a
                    # predicted output of a non-skipped quantum.
                    skeleton[output_dataset_key]["ref"] = output_ref
                else:
                    # Remove this dataset from the skeleton graph,
                    # because the quantum that would have produced it
                    # is being skipped and it doesn't already exist.
                    skeleton.remove_dataset_nodes([output_dataset_key])
                # If this dataset was "in the way" (i.e. already in the
                # output run), it isn't anymore.
                self.existing_datasets.outputs_in_the_way.pop(output_dataset_key, None)
            # Removing the quantum node from the graph will happen outside this
            # function.
            return True
        return False

    @final
    def _gather_quantum_outputs(
        self,
        task_node: TaskNode,
        quantum_key: QuantumKey,
        skeleton: QuantumGraphSkeleton,
        skypix_bounds_builder: SkyPixBoundsBuilder,
        timespan_builder: TimespanBuilder,
    ) -> NamedKeyDict[DatasetType, list[DatasetRef]]:
        """Collect outputs or generate datasets for a preliminary quantum and
        put them in the form used by `~lsst.daf.butler.Quantum` and
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.
        skypix_bounds_builder : `~prerequisite_helpers.SkyPixBoundsBuilder`
            An object that accumulates the appropriate spatial bounds for a
            quantum.
        timespan_builder : `~prerequisite_helpers.TimespanBuilder`
            An object that accumulates the appropriate timespan for a quantum.

        Returns
        -------
        outputs : `~lsst.daf.butler.NamedKeyDict` [ \
                `~lsst.daf.butler.DatasetType`, `list` [ \
                `~lsst.daf.butler.DatasetRef` ] ]
            All outputs to the task, using the storage class and components
            defined by the task's own connections.

        Notes
        -----
        This first looks for outputs already present in the `output_run` by
        looking in `ExistingDatasets.outputs_in_the_way`; if it finds something
        and `clobber` is `True`, it uses that ref (it's not ideal that both the
        original dataset and its replacement will have the same UUID, but we
        don't have space in the quantum graph for two UUIDs, and we need the
        datastore records of the original there).  If `clobber` is `False`,
        `RuntimeError` is raised.  If there is no output already present, a new
        one with a random UUID is generated.  In all cases the "ref" attribute
        of the dataset node in the skeleton is set.
        """
        outputs_by_type: dict[str, list[DatasetRef]] = {}
        dataset_key: DatasetKey
        for dataset_key in skeleton.iter_outputs_of(quantum_key):
            dataset_data_id = skeleton[dataset_key]["data_id"]
            dataset_type_node = self._pipeline_graph.dataset_types[dataset_key.parent_dataset_type_name]
            if (ref := self.existing_datasets.outputs_in_the_way.get(dataset_key)) is None:
                ref = DatasetRef(dataset_type_node.dataset_type, dataset_data_id, run=self.output_run)
            elif not self.clobber:
                # We intentionally raise here, before running adjustQuantum,
                # because it'd be weird if we left an old potential output of a
                # task sitting there in the output collection, just because the
                # task happened to not actually produce it.
                raise OutputExistsError(
                    f"Potential output dataset {ref} already exists in the output run "
                    f"{self.output_run}, but clobbering outputs was not expected to be necessary."
                )
            skypix_bounds_builder.handle_dataset(dataset_key.parent_dataset_type_name, dataset_data_id)
            timespan_builder.handle_dataset(dataset_key.parent_dataset_type_name, dataset_data_id)
            skeleton[dataset_key]["ref"] = ref
            outputs_by_type.setdefault(dataset_key.parent_dataset_type_name, []).append(ref)
        adapted_outputs: NamedKeyDict[DatasetType, list[DatasetRef]] = NamedKeyDict()
        for write_edge in task_node.iter_all_outputs():
            dataset_type_node = self._pipeline_graph.dataset_types[write_edge.parent_dataset_type_name]
            edge_dataset_type = write_edge.adapt_dataset_type(dataset_type_node.dataset_type)
            adapted_outputs[edge_dataset_type] = [
                write_edge.adapt_dataset_ref(ref)
                for ref in sorted(outputs_by_type.get(write_edge.parent_dataset_type_name, []))
            ]
        return adapted_outputs

    @final
    def _gather_quantum_inputs(
        self,
        task_node: TaskNode,
        quantum_key: QuantumKey,
        skeleton: QuantumGraphSkeleton,
        task_prerequisite_info: PrerequisiteInfo,
        skypix_bounds_builder: SkyPixBoundsBuilder,
        timespan_builder: TimespanBuilder,
    ) -> NamedKeyDict[DatasetType, list[DatasetRef]]:
        """Collect input datasets for a preliminary quantum and put them in the
        form used by `~lsst.daf.butler.Quantum` and
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Node for this task in the pipeline graph.
        quantum_key : `QuantumKey`
            Identifier for this quantum in the graph.
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph, to be modified in-place.
        skypix_bounds_builder : `~prerequisite_helpers.SkyPixBoundsBuilder`
            An object that accumulates the appropriate spatial bounds for a
            quantum.
        timespan_builder : `~prerequisite_helpers.TimespanBuilder`
            An object that accumulates the appropriate timespan for a quantum.

        Returns
        -------
        inputs : `~lsst.daf.butler.NamedKeyDict` [ \
                `~lsst.daf.butler.DatasetType`, `list` [ \
                `~lsst.daf.butler.DatasetRef` ] ]
            All regular and prerequisite inputs to the task, using the storage
            class and components defined by the task's own connections.

        Notes
        -----
        On return, the dataset nodes that represent inputs to this quantum will
        either have their "ref" attribute set (using the common dataset type,
        not the task-specific one) or will be removed from the graph.

        For regular inputs, usually an existing "ref" (corresponding to an
        output of another quantum) will be found and left unchanged.  When
        there is no existing "ref" attribute, `ExistingDatasets.inputs` is
        searched next; if there is nothing there, the input will be removed.

        Prerequisite inputs are always queried for directly here (delegating to
        `_find_prerequisite_inputs`).  They are never produced by other tasks,
        and cannot in general be queried for in advance when
        `ExistingDatasets.inputs` is populated.
        """
        quantum_data_id = skeleton[quantum_key]["data_id"]
        inputs_by_type: dict[str, set[DatasetRef]] = {}
        dataset_key: DatasetKey | PrerequisiteDatasetKey
        # Process inputs already present in the skeleton - this should include
        # all regular inputs (including intermediates) and may include some
        # prerequisites.
        for dataset_key in list(skeleton.iter_inputs_of(quantum_key)):
            if (ref := skeleton[dataset_key].get("ref")) is None:
                # This dataset is an overall input - if it was an intermediate,
                # we would have already either removed the node or set the
                # "ref" attribute when processing its producing quantum - and
                # this is the first time we're trying to resolve it.
                if (ref := self.existing_datasets.inputs.get(dataset_key)) is None:
                    # It also doesn't exist in the input collections, so we
                    # remove its node in the skeleton graph (so other consumers
                    # won't have to check for it).
                    skeleton.remove_dataset_nodes([dataset_key])
                    continue
                skeleton[dataset_key]["ref"] = ref
            inputs_by_type.setdefault(dataset_key.parent_dataset_type_name, set()).add(ref)
            skypix_bounds_builder.handle_dataset(dataset_key.parent_dataset_type_name, ref.dataId)
            timespan_builder.handle_dataset(dataset_key.parent_dataset_type_name, ref.dataId)
        # Query for any prerequisites not handled by process_subgraph.  Note
        # that these were not already in the skeleton graph, so we add them
        # now.
        skypix_bounds = skypix_bounds_builder.finish()
        timespan = timespan_builder.finish()
        for finder in task_prerequisite_info.finders.values():
            inputs_for_type = inputs_by_type.setdefault(finder.dataset_type_node.name, set())
            dataset_keys = []
            for ref in finder.find(
                self.butler, self.input_collections, quantum_data_id, skypix_bounds, timespan
            ):
                dataset_key = skeleton.add_prerequisite_node(ref.datasetType.name, ref=ref)
                dataset_keys.append(dataset_key)
                inputs_for_type.add(ref)
            skeleton.add_input_edges(quantum_key, dataset_keys)
        adapted_inputs: NamedKeyDict[DatasetType, list[DatasetRef]] = NamedKeyDict()
        for read_edge in task_node.iter_all_inputs():
            dataset_type_node = self._pipeline_graph.dataset_types[read_edge.parent_dataset_type_name]
            edge_dataset_type = read_edge.adapt_dataset_type(dataset_type_node.dataset_type)
            if (current_dataset_type := adapted_inputs.keys().get(edge_dataset_type.name)) is None:
                adapted_inputs[edge_dataset_type] = [
                    read_edge.adapt_dataset_ref(ref)
                    for ref in sorted(inputs_by_type.get(read_edge.parent_dataset_type_name, frozenset()))
                ]
            elif current_dataset_type != edge_dataset_type:
                raise NotImplementedError(
                    f"Task {task_node.label!r} has {edge_dataset_type.name!r} as an input via "
                    "two different connections, with two different storage class overrides. "
                    "This is not yet supported due to limitations in the Quantum data structure."
                )
            # If neither the `if` nor the `elif` above match, it means
            # multiple input connections have exactly the same dataset
            # type, and hence nothing to do after the first one.
        return adapted_inputs

    @final
    def _resolve_task_init(
        self, task_node: TaskNode, skeleton: QuantumGraphSkeleton, has_skipped_quanta: bool
    ) -> None:
        """Add init-input and init-output dataset nodes and edges for a task to
        the skeleton.

        Parameters
        ----------
        task_node : `pipeline_graph.TaskNode`
            Pipeline graph description of the task.
        skeleton : `QuantumGraphSkeleton`
            In-progress quantum graph data structure to update in-place.
        has_skipped_quanta : `bool`
            Whether any of this task's quanta were skipped because they had
            already succeeded.
        """
        quanta = skeleton.get_quanta(task_node.label)
        task_init_key = TaskInitKey(task_node.label)
        if quanta:
            adapted_inputs: NamedKeyDict[DatasetType, DatasetRef] = NamedKeyDict()
            # Process init-inputs.
            input_keys: list[DatasetKey] = []
            for read_edge in task_node.init.iter_all_inputs():
                dataset_key = skeleton.add_dataset_node(
                    read_edge.parent_dataset_type_name, self.empty_data_id
                )
                skeleton.add_input_edge(task_init_key, dataset_key)
                if (ref := skeleton[dataset_key].get("ref")) is None:
                    try:
                        ref = self.existing_datasets.inputs[dataset_key]
                    except KeyError:
                        raise InitInputMissingError(
                            f"Overall init-input dataset {read_edge.parent_dataset_type_name!r} "
                            f"needed by task {task_node.label!r} not found in input collection(s) "
                            f"{self.input_collections}."
                        ) from None
                    skeleton[dataset_key]["ref"] = ref
                for quantum_key in skeleton.get_quanta(task_node.label):
                    skeleton.add_input_edge(quantum_key, dataset_key)
                input_keys.append(dataset_key)
                adapted_ref = read_edge.adapt_dataset_ref(ref)
                adapted_inputs[adapted_ref.datasetType] = adapted_ref
            # Save the quantum-adapted init inputs to each quantum, and add
            # skeleton edges connecting the init inputs to each quantum.
            for quantum_key in skeleton.get_quanta(task_node.label):
                skeleton[quantum_key]["init_inputs"] = adapted_inputs
            # Process init-outputs.
            adapted_outputs: NamedKeyDict[DatasetType, DatasetRef] = NamedKeyDict()
            for write_edge in task_node.init.iter_all_outputs():
                dataset_key = skeleton.add_dataset_node(
                    write_edge.parent_dataset_type_name, self.empty_data_id
                )
                if (ref := self.existing_datasets.outputs_in_the_way.get(dataset_key)) is None:
                    ref = DatasetRef(
                        self._pipeline_graph.dataset_types[write_edge.parent_dataset_type_name].dataset_type,
                        self.empty_data_id,
                        run=self.output_run,
                    )
                skeleton[dataset_key]["ref"] = ref
                skeleton.add_output_edge(task_init_key, dataset_key)
                adapted_ref = write_edge.adapt_dataset_ref(ref)
                adapted_outputs[adapted_ref.datasetType] = adapted_ref
            skeleton[task_init_key]["inputs"] = adapted_inputs
            skeleton[task_init_key]["outputs"] = adapted_outputs
        elif has_skipped_quanta:
            # No quanta remain for this task, but at least one quantum was
            # skipped because its outputs were present in the skip_existing_in
            # collections.  This means all init outputs should be present in
            # the skip_existing_in collections, too, and we need to put those
            # refs in the graph.
            for write_edge in task_node.init.iter_all_outputs():
                dataset_key = skeleton.add_dataset_node(
                    write_edge.parent_dataset_type_name, self.empty_data_id
                )
                if (ref := self.existing_datasets.outputs_for_skip.get(dataset_key)) is None:
                    raise InitInputMissingError(
                        f"Init-output dataset {write_edge.parent_dataset_type_name!r} of skipped task "
                        f"{task_node.label!r} not found in skip-existing-in collection(s) "
                        f"{self.skip_existing_in}."
                    ) from None
                skeleton[dataset_key]["ref"] = ref
                # If this dataset was "in the way" (i.e. already in the output
                # run), it isn't anymore.
                self.existing_datasets.outputs_in_the_way.pop(dataset_key, None)
        # No quanta remain in this task, but none were skipped; this means
        # they all got pruned because of NoWorkFound conditions.  This
        # dooms all downstream quanta to the same fate, so we don't bother
        # doing anything with the task's init-outputs, since nothing is
        # going to consume them.

    @final
    @timeMethod
    def _find_empty_dimension_datasets(self) -> None:
        """Query for all dataset types with no dimensions, updating
        `existing_datasets` in-place.

        This includes but is not limited to init inputs and init outputs.
        """
        _, dataset_type_nodes = self._pipeline_graph.group_by_dimensions()[self.universe.empty.as_group()]
        dataset_types = [node.dataset_type for node in dataset_type_nodes.values()]
        dataset_types.extend(self._global_init_output_types.values())
        for dataset_type in dataset_types:
            key = DatasetKey(dataset_type.name, self.empty_data_id.required_values)
            if (
                self._pipeline_graph.producer_of(dataset_type.name) is None
                and dataset_type.name not in self._global_init_output_types
            ):
                # Dataset type is an overall input; we always need to try to
                # find these.
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=self.input_collections)
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    self.existing_datasets.inputs[key] = ref
            elif self.skip_existing_in:
                # Dataset type is an intermediate or output; need to find these
                # if only they're from previously executed quanta that we might
                # skip...
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=self.skip_existing_in)
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    self.existing_datasets.outputs_for_skip[key] = ref
                    if ref.run == self.output_run:
                        self.existing_datasets.outputs_in_the_way[key] = ref
            if self.output_run_exists and not self.skip_existing_starts_with_output_run:
                # ...or if they're in the way and would need to be clobbered
                # (and we haven't already found them in the previous block).
                try:
                    ref = self.butler.find_dataset(dataset_type.name, collections=[self.output_run])
                except MissingDatasetTypeError:
                    ref = None
                if ref is not None:
                    self.existing_datasets.outputs_in_the_way[key] = ref

    @final
    @timeMethod
    def _attach_datastore_records(self, skeleton: QuantumGraphSkeleton) -> None:
        """Add datastore records for all overall inputs to a preliminary
        quantum graph.

        Parameters
        ----------
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph to update in place.

        Notes
        -----
        On return, all quantum nodes in the skeleton graph will have a
        "datastore_records" attribute that is a mapping from datastore name
        to `lsst.daf.butler.DatastoreRecordData`, as used by
        `lsst.daf.butler.Quantum`.
        """
        overall_inputs = skeleton.extract_overall_inputs()
        exported_records = self.butler._datastore.export_records(overall_inputs.values())
        for quantum_key in skeleton.iter_all_quanta():
            quantum_records = {}
            input_ids = {
                ref.id
                for dataset_key in skeleton.iter_inputs_of(quantum_key)
                if (ref := overall_inputs.get(dataset_key)) is not None
            }
            if input_ids:
                for datastore_name, records in exported_records.items():
                    matching_records = records.subset(input_ids)
                    if matching_records is not None:
                        quantum_records[datastore_name] = matching_records
            skeleton[quantum_key]["datastore_records"] = quantum_records

    @final
    @timeMethod
    def _construct_quantum_graph(
        self, skeleton: QuantumGraphSkeleton, metadata: Mapping[str, Any]
    ) -> QuantumGraph:
        """Construct a `QuantumGraph` object from the contents of a
        fully-processed `quantum_graph_skeleton.QuantumGraphSkeleton`.

        Parameters
        ----------
        skeleton : `quantum_graph_skeleton.QuantumGraphSkeleton`
            Preliminary quantum graph.  Must have "init_inputs", "inputs", and
            "outputs" attributes on all quantum nodes, as added by
            `_resolve_task_quanta`, as well as a "datastore_records" attribute
            as added by `_attach_datastore_records`.
        metadata : `Mapping`
            Flexible metadata to add to the graph.

        Returns
        -------
        quantum_graph : `QuantumGraph`
            DAG describing processing to be performed.
        """
        quanta: dict[TaskDef, set[Quantum]] = {}
        init_inputs: dict[TaskDef, Iterable[DatasetRef]] = {}
        init_outputs: dict[TaskDef, Iterable[DatasetRef]] = {}
        for task_def in self._pipeline_graph._iter_task_defs():
            if not skeleton.has_task(task_def.label):
                continue
            task_node = self._pipeline_graph.tasks[task_def.label]
            task_init_key = skeleton.get_task_init_node(task_def.label)
            init_inputs[task_def] = skeleton[task_init_key]["inputs"].values()
            init_outputs[task_def] = skeleton[task_init_key]["outputs"].values()
            quanta_for_task: set[Quantum] = set()
            for quantum_key in skeleton.get_quanta(task_node.label):
                node_state = skeleton[quantum_key]
                quanta_for_task.add(
                    Quantum(
                        taskName=task_node.task_class_name,
                        taskClass=task_node.task_class,
                        dataId=node_state["data_id"],
                        initInputs=node_state["init_inputs"],
                        inputs=node_state["inputs"],
                        outputs=node_state["outputs"],
                        datastore_records=node_state.get("datastore_records"),
                    )
                )
            quanta[task_def] = quanta_for_task

        registry_dataset_types: list[DatasetType] = [
            node.dataset_type for node in self._pipeline_graph.dataset_types.values()
        ]

        all_metadata = self.metadata.to_dict()
        all_metadata.update(metadata)
        return QuantumGraph(
            quanta,
            metadata=all_metadata,
            universe=self.universe,
            initInputs=init_inputs,
            initOutputs=init_outputs,
            globalInitOutputs=[skeleton[key]["ref"] for key in skeleton.global_init_outputs],
            registryDatasetTypes=registry_dataset_types,
        )

    @staticmethod
    @final
    def _find_removed(
        original: Iterable[DatasetKey | PrerequisiteDatasetKey],
        adjusted: NamedKeyMapping[DatasetType, Sequence[DatasetRef]],
    ) -> set[DatasetKey | PrerequisiteDatasetKey]:
        """Identify skeleton-graph dataset nodes that have been removed by
        `~PipelineTaskConnections.adjustQuantum`.

        Parameters
        ----------
        original : `~collections.abc.Iterable` [ `DatasetKey` or \
                `PrerequisiteDatasetKey` ]
            Identifiers for the dataset nodes that were the original neighbors
            (inputs or outputs) of a quantum.
        adjusted : `~lsst.daf.butler.NamedKeyMapping` [ \
                `~lsst.daf.butler.DatasetType`, \
                `~collections.abc.Sequence` [ `lsst.daf.butler.DatasetType` ] ]
            Adjusted neighbors, in the form used by `lsst.daf.butler.Quantum`.

        Returns
        -------
        removed : `set` [ `DatasetKey` ]
            Datasets in ``original`` that have no counterpart in ``adjusted``.
        """
        result = set(original)
        for dataset_type, kept_refs in adjusted.items():
            parent_dataset_type_name, _ = DatasetType.splitDatasetTypeName(dataset_type.name)
            for kept_ref in kept_refs:
                # We don't know if this was a DatasetKey or a
                # PrerequisiteDatasetKey; just try both.
                result.discard(DatasetKey(parent_dataset_type_name, kept_ref.dataId.required_values))
                result.discard(PrerequisiteDatasetKey(parent_dataset_type_name, kept_ref.id.bytes))
        return result


@dataclasses.dataclass(eq=False, order=False)
class ExistingDatasets:
    """Struct that holds the results of dataset queries for
    `QuantumGraphBuilder`.
    """

    inputs: dict[DatasetKey | PrerequisiteDatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Overall-input datasets found in `QuantumGraphBuilder.input_collections`.

    This may include prerequisite inputs.  It does include init-inputs.
    It does not include intermediates.
    """

    outputs_for_skip: dict[DatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Output datasets found in `QuantumGraphBuilder.skip_existing_in`.

    It is unspecified whether this contains include init-outputs; there is
    no concept of skipping at the init stage, so this is not expected to
    matter.
    """

    outputs_in_the_way: dict[DatasetKey, DatasetRef] = dataclasses.field(default_factory=dict)
    """Output datasets found in `QuantumGraphBuilder.output_run`.

    This includes regular outputs and init-outputs.
    """


def _quantum_or_quanta(n: int) -> str:
    """Correctly pluralize 'quantum' if needed."""
    return f"{n} quanta" if n != 1 else "1 quantum"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("TaskSubset",)

from collections.abc import Iterator, MutableSet

import networkx
import networkx.algorithms.boundary

from ._exceptions import PipelineGraphError
from ._nodes import NodeKey, NodeType


class TaskSubset(MutableSet[str]):
    """A specialized set that represents a labeles subset of the tasks in a
    pipeline graph.

    Instances of this class should never be constructed directly; they should
    only be accessed via the `PipelineGraph.task_subsets` attribute and created
    by the `PipelineGraph.add_task_subset` method.

    Parameters
    ----------
    parent_xgraph : `networkx.DiGraph`
        Parent networkx graph that this subgraph is part of.
    label : `str`
        Label associated with this subset of the pipeline.
    members : `set` [ `str` ]
        Labels of the tasks that are members of this subset.
    description : `str`, optional
        Description string associated with this labeled subset.

    Notes
    -----
    Iteration order is arbitrary, even when the parent pipeline graph is
    ordered (there is no guarantee that an ordering of the tasks in the graph
    implies a consistent ordering of subsets).
    """

    def __init__(
        self,
        parent_xgraph: networkx.DiGraph,
        label: str,
        members: set[str],
        description: str,
    ):
        self._parent_xgraph = parent_xgraph
        self._label = label
        self._members = members
        self._description = description

    @property
    def label(self) -> str:
        """Label associated with this subset of the pipeline."""
        return self._label

    @property
    def description(self) -> str:
        """Description string associated with this labeled subset."""
        return self._description

    @description.setter
    def description(self, value: str) -> None:
        # Docstring in getter.
        self._description = value

    def __repr__(self) -> str:
        return f"{self.label}: {self.description!r}, tasks={{{', '.join(iter(self))}}}"

    def __contains__(self, key: object) -> bool:
        return key in self._members

    def __len__(self) -> int:
        return len(self._members)

    def __iter__(self) -> Iterator[str]:
        return iter(self._members)

    def add(self, task_label: str) -> None:
        """Add a new task to this subset.

        Parameters
        ----------
        task_label : `str`
            Label for the task.  Must already be present in the parent pipeline
            graph.
        """
        key = NodeKey(NodeType.TASK, task_label)
        if key not in self._parent_xgraph:
            raise PipelineGraphError(f"{task_label!r} is not a task in the parent pipeline.")
        self._members.add(key.name)

    def discard(self, task_label: str) -> None:
        """Remove a task from the subset if it is present.

        Parameters
        ----------
        task_label : `str`
            Label for the task.  Must already be present in the parent pipeline
            graph.
        """
        self._members.discard(task_label)



#
# LSST Data Management System
# Copyright 2008-2015 AURA/LSST.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <https://www.lsstcorp.org/LegalNotices/>.
#
from __future__ import annotations

__all__: list[str] = []

import contextlib
import logging
from collections.abc import Generator
from typing import TYPE_CHECKING

from deprecated.sphinx import deprecated

if TYPE_CHECKING:
    import cProfile


@deprecated(
    reason="Replaced by lsst.utils.timer.profile().  Will be removed after v26.0",
    version="v25.0",
    category=FutureWarning,
)
@contextlib.contextmanager
def profile(filename: str, log: logging.Logger | None = None) -> Generator[cProfile.Profile, None, None]:
    """Context manager for profiling with cProfile.

    Parameters
    ----------
    filename : `str`
        Filename to which to write profile (profiling disabled if `None` or
        empty).
    log : `logging.Logger`, optional
        Log object for logging the profile operations.

    Yields
    ------
    `cProfile.Profile`
        Profiling object.

    Notes
    -----
    If profiling is enabled, the context manager returns the cProfile.Profile
    object (otherwise it returns None), which allows additional control over
    profiling.  You can obtain this using the "as" clause, e.g.:

    .. code-block:: python

        with profile(filename) as prof:
            runYourCodeHere()

    The output cumulative profile can be printed with a command-line like:

    .. code-block:: bash

        python -c 'import pstats; \
            pstats.Stats("<filename>").sort_stats("cumtime").print_stats(30)'
    """
    if not filename:
        # Nothing to do
        yield None  # type: ignore
        return
    from cProfile import Profile

    profile = Profile()
    if log is not None:
        log.info("Enabling cProfile profiling")
    profile.enable()
    yield profile
    profile.disable()
    profile.dump_stats(filename)
    if log is not None:
        log.info("cProfile stats written to %s", filename)
